const HELP_SYMBOL = Symbol('hxs-help-symbol');
const CONSTRUCTOR_SYMBOL = Symbol('hxs-constructor-symbol');
const BASE_SYMBOL = Symbol('hxs-base-symbol');
var Utils;
(function (Utils) {
    /**
     * A stable in-place sorting function.
     */
    Utils.sort = (array, compare) => {
        let t;
        for (let i = 0; i < array.length - 1; i++) {
            for (let j = i + 1; j < array.length; j++) {
                if (compare(array[i], array[j]) > 0) {
                    t = array[i];
                    array[i] = array[j];
                    array[j] = t;
                }
            }
        }
    };
    /**
     * Remove specific element(s) from the array.
     * (Default count: `array.length - start`)
     */
    Utils.removeElements = (array, start, count = array.length - start) => {
        for (let i = start; i < array.length - count; i++) {
            array[i] = array[i + count];
        }
        array.length -= count;
    };
    Utils.normalizeIndex = (index, arraySize, referrer, context, allowEnd = false) => {
        const normalizedIndex = index < 0
            ? arraySize + index
            : index;
        if (normalizedIndex > arraySize
            || normalizedIndex < 0
            || (!allowEnd && normalizedIndex === arraySize)) {
            const message = `index(${index}) out of range(array size: ${arraySize})`;
            Utils.raise(RangeError, message, referrer, context);
        }
        return normalizedIndex;
    };
    Utils.formatFrameString = (referrer, context) => (`  at ${context.source} (Ln ${referrer.line}, Col ${referrer.column})`);
    /**
     * Throw a specific error with environment info appended.
     */
    Utils.raise = (constructor, message, referrer, context) => {
        const currentFrame = Utils.formatFrameString(referrer, context);
        const { stack } = context;
        let errorMessage = message;
        if (stack[stack.length - 1] !== currentFrame) {
            errorMessage += '\n' + currentFrame;
        }
        for (let i = stack.length - 1; i >= 0; i--) {
            errorMessage += '\n' + stack[i];
        }
        throw new constructor(errorMessage);
    };
    Utils.parseNumber = (node, context) => {
        let result;
        switch (node.suffix) {
            case 'D':
            case '': {
                result = +node.value;
                break;
            }
            case 'B': {
                result = Number.parseInt(node.value, 2);
                break;
            }
            case 'O': {
                result = Number.parseInt(node.value, 8);
                break;
            }
            case 'H': {
                result = Number.parseInt(node.value, 16);
                break;
            }
            default: {
                Utils.raise(SyntaxError, 'unrecognized number', node, context);
            }
        }
        if (result !== result) { // NaN
            Utils.raise(SyntaxError, 'invalid number', node, context);
        }
        return result;
    };
    Utils.isDict = (value) => (Object.prototype.toString.call(value) === '[object Object]');
    Utils.createDict = (dict) => (Object.assign(Object.create(null), dict));
    Utils.createValueNode = (value, referrer) => ({
        type: 'value',
        line: referrer.line,
        column: referrer.column,
        offset: referrer.offset,
        value,
    });
    Utils.createRawArray = (array, referrer) => {
        const COMMA_NODE = {
            type: 'symbol',
            value: ',',
            line: referrer.line,
            column: referrer.column,
            offset: referrer.offset,
        };
        const rawArray = [];
        for (let i = 0; i < array.length; i++) {
            if (i) {
                rawArray.push(COMMA_NODE);
            }
            rawArray.push(Utils.createValueNode(array[i], referrer));
        }
        return rawArray;
    };
    Utils.replaceBuffer = (buffer, start, width, replacement) => {
        buffer[start] = replacement;
        if (width > 1) {
            Utils.removeElements(buffer, start + 1, width - 1);
        }
    };
    /**
     * Excute the callback with the temporary variables
     * defined in context store and recover conflicting
     * variables after execution.
     */
    Utils.injectTemp = (store, variables, callback) => {
        const copy = Object.create(null);
        for (const key in variables) {
            if (key in store) {
                copy[key] = store[key];
            }
            store[key] = variables[key];
        }
        callback();
        for (const key in variables) {
            if (key in copy) {
                store[key] = copy[key];
            }
            else {
                delete store[key];
            }
        }
    };
    Utils.extendContext = (context, store) => ({
        store,
        exports: context.exports,
        resolvedModules: context.resolvedModules,
        source: context.source,
        basePath: context.basePath,
        stack: context.stack,
    });
    Utils.injectHelp = (helpInfo, target) => {
        target[HELP_SYMBOL] = helpInfo;
        return target;
    };
    Utils.toString = (value, previewSize = 10, previewIndent = '  ') => {
        const type = typeof value;
        if (type === 'string') {
            return value;
        }
        else if (type === 'function') {
            return '<function>';
        }
        else if (Array.isArray(value)) {
            if (previewSize <= 0) {
                return '<array>';
            }
            let result = `(size: ${value.length}) [`;
            if (value.length > 0) {
                result += '\n' + previewIndent;
                result += value.slice(0, previewSize)
                    .map(v => (Utils.toDisplay(v, 0) + ',\n'))
                    .join(previewIndent);
            }
            if (value.length > previewSize) {
                result += previewIndent + '...\n';
            }
            result += ']';
            return result;
        }
        else if (Utils.isDict(value)) {
            if (previewSize <= 0) {
                return '<dict>';
            }
            const entries = Object.entries(value);
            let result = `(size: ${entries.length}) {`;
            if (entries.length > 0) {
                result += '\n' + previewIndent;
                result += entries.slice(0, previewSize)
                    .map(entry => (Utils.toDisplay(entry[0], 0)
                    + ' -> '
                    + Utils.toDisplay(entry[1], 0)
                    + ',\n'))
                    .join(previewIndent);
                if (entries.length > previewSize) {
                    result += previewIndent + '...\n';
                }
            }
            result += '}';
            return result;
        }
        else {
            return String(value);
        }
    };
    Utils.toDisplay = (value, previewSize, previewIndent) => {
        if (typeof value !== 'string') {
            return Utils.toString(value, previewSize, previewIndent);
        }
        if (!value.includes("'")) {
            return "'" + value + "'";
        }
        else if (!value.includes('"')) {
            return '"' + value + '"';
        }
        else if (!value.includes('`')) {
            return '`' + value + '`';
        }
        else {
            return "'" + value.replace(/'/g, "\\'") + "'";
        }
    };
})(Utils || (Utils = {}));

const getOperatorSymbol = (node) => (node.type === 'symbol' ? node.value : node.start);

/**
 * Evalute a single node.
 */
const evalNode = (node, context) => {
    switch (node.type) {
        case 'value': {
            return node.value;
        }
        case 'glob': {
            return node.value.slice(1, -1);
        }
        case 'number': {
            return Utils.parseNumber(node, context);
        }
        case 'span': {
            return evalExpression(node.body, context);
        }
        case 'word': {
            const name = node.value;
            const { store } = context;
            if (!(name in store)) {
                Utils.raise(ReferenceError, `${Utils.toDisplay(name)} is not defined`, node, context);
            }
            return store[name];
        }
        default: {
            Utils.raise(SyntaxError, 'unrecognized syntax', node, context);
            return null; // for type checking (ts v4.3.4)
        }
    }
};

const compileExpression = (buffer, context) => {
    let operatorNodes = [];
    // find all operators (and spans)
    for (let i = 0; i < buffer.length; i++) {
        const node = buffer[i];
        if (node.type === 'symbol') {
            operatorNodes.push(node);
            if (!operatorPriorities.has(node.value)) {
                Utils.raise(SyntaxError, `unknown operator ${Utils.toDisplay(node.value)}`, node, context);
            }
        }
        else if (node.type === 'span') {
            if (operatorPriorities.has(node.start)) {
                operatorNodes.push(node);
            }
        }
    }
    // sort operators in order of priority (high->low)
    Utils.sort(operatorNodes, (nodeA, nodeB) => {
        const priorityA = operatorPriorities.get(getOperatorSymbol(nodeA));
        const priorityB = operatorPriorities.get(getOperatorSymbol(nodeB));
        return priorityA - priorityB;
    });
    let p = 0;
    let t;
    while (p < operatorNodes.length) {
        let symbol = getOperatorSymbol(operatorNodes[p]);
        for (let q = p + 1; q <= operatorNodes.length; q++) {
            if (q < operatorNodes.length
                && getOperatorSymbol(operatorNodes[q]) === symbol) {
                continue;
            }
            if (operatorLtr.get(symbol)) {
                p = q;
                break;
            }
            let l = p;
            let r = q - 1;
            while (l < r) {
                t = operatorNodes[l];
                operatorNodes[l] = operatorNodes[r];
                operatorNodes[r] = t;
                l++;
                r--;
            }
            p = q;
        }
    }
    return { buffer, operatorNodes };
};
const evalCompiledExpression = (compiledExpression, context, copyBuffers, optimizeOperators) => {
    const { operatorNodes, buffer: _buffer } = compiledExpression;
    const buffer = copyBuffers ? _buffer.slice() : _buffer;
    let extraOperatorIndices = null;
    if (optimizeOperators) {
        extraOperatorIndices = new Set();
    }
    // execute operators
    for (let i = 0; i < operatorNodes.length; i++) {
        const operatorNode = operatorNodes[i];
        const index = buffer.indexOf(operatorNode);
        if (~index) { // index !== -1
            const handler = operatorHandlers.get(operatorNode.type === 'symbol'
                ? operatorNode.value
                : operatorNode.start);
            handler(buffer, index, context);
        }
        else if (optimizeOperators) {
            extraOperatorIndices.add(i);
        }
    }
    // remove extra operators
    if (extraOperatorIndices) {
        const optimizedOperators = operatorNodes.filter((_, i) => !extraOperatorIndices.has(i));
        if (optimizedOperators.length < operatorNodes.length) {
            for (let i = 0; i < optimizedOperators.length; i++) {
                operatorNodes[i] = optimizedOperators[i];
            }
            operatorNodes.length = optimizedOperators.length;
        }
    }
    // check result
    if (buffer.length > 1) {
        Utils.raise(SyntaxError, 'unrecognized syntax', buffer[1], context);
        return null; // for type checking
    }
    else if (buffer.length === 0) {
        return null; // for type checking
    }
    else {
        return evalNode(buffer[0], context);
    }
};
/**
 * Evalute an expression.
 */
const evalExpression = (nodes, context, begin = 0, end = nodes.length) => {
    if (begin >= end) {
        return null;
    }
    const buffer = nodes.slice(begin, end);
    return evalCompiledExpression(compileExpression(buffer, context), context, false, false);
};

const compileNodes = (nodes, context, begin = 0, end = nodes.length) => {
    const compiledExpressions = [];
    if (begin >= end) {
        return {
            compiledExpressions,
            endsWithSemicolon: true,
        };
    }
    let left = begin;
    let right;
    for (right = 0; right <= end; right++) {
        if (right < end) {
            const node = nodes[right];
            if (node.type !== 'symbol' || node.value !== ';') {
                continue;
            }
        }
        const buffer = nodes.slice(left, right);
        compiledExpressions.push(compileExpression(buffer, context));
        if (right < end - 1) {
            left = right + 1;
        }
        else {
            break;
        }
    }
    const endsWithSemicolon = right < end;
    return { compiledExpressions, endsWithSemicolon };
};
const evalCompiledNodes = (compiledNodes, context, copyBuffers, optimizeOperators) => {
    const { compiledExpressions, endsWithSemicolon } = compiledNodes;
    let lastValue = null;
    for (let i = 0; i < compiledExpressions.length; i++) {
        lastValue = evalCompiledExpression(compiledExpressions[i], context, copyBuffers, optimizeOperators);
        context.store._ = lastValue;
    }
    return endsWithSemicolon ? null : lastValue;
};
/**
 * Evaluate the given nodes.
 */
const evalNodes = (nodes, context, begin = 0, end = nodes.length) => (evalCompiledNodes(compileNodes(nodes, context, begin, end), context, false, false));

/**
 * Parse single argument declaration.
 */
const parseArg = (rawArgs, begin, end, context, allowRequired) => {
    const firstNode = rawArgs[begin];
    const secondNode = rawArgs[begin + 1];
    if (!firstNode || firstNode.type !== 'word') {
        Utils.raise(SyntaxError, 'expect a word as argument name', firstNode, context);
    }
    if (end - begin === 1) {
        if (!allowRequired) {
            Utils.raise(SyntaxError, 'required argument must be before optional ones', firstNode, context);
        }
        return {
            name: firstNode.value,
            required: true,
            rest: false,
            default: null,
        };
    }
    else if (end - begin === 2) {
        if (secondNode.type !== 'symbol'
            || (secondNode.value !== '...' && secondNode.value !== '?')) {
            Utils.raise(SyntaxError, 'invalid argument declaration', firstNode, context);
        }
        return {
            name: firstNode.value,
            required: false,
            rest: secondNode.value === '...',
            default: null,
        };
    }
    else {
        if (secondNode.type !== 'symbol' || secondNode.value !== '=') {
            Utils.raise(SyntaxError, 'invalid argument declaration', firstNode, context);
        }
        return {
            name: firstNode.value,
            required: false,
            rest: false,
            default: compileExpression(rawArgs.slice(begin + 2, end), context),
        };
    }
};
/**
 * Parse argument list.
 */
const parseArgList = (rawArgList, context) => {
    if (rawArgList.length === 0) {
        return {
            args: [],
            requiredCount: 0,
            restArg: null,
        };
    }
    const args = [];
    let allowRequired = true;
    let requiredCount = 0;
    let l = 0;
    let argDef;
    let restArg = null;
    for (let r = 0; r <= rawArgList.length; r++) {
        if (r < rawArgList.length) {
            const node = rawArgList[r];
            if (node.type !== 'symbol' || node.value !== ',') {
                continue;
            }
        }
        if (restArg !== null) {
            Utils.raise(SyntaxError, 'rest argument must be the last one', rawArgList[l], context);
        }
        argDef = parseArg(rawArgList, l, r, context, allowRequired);
        if (argDef.rest) {
            restArg = argDef.name;
        }
        else {
            args.push(argDef);
            if (argDef.required) {
                requiredCount++;
            }
            else {
                allowRequired = false;
            }
        }
        if (r < rawArgList.length - 1) {
            l = r + 1;
        }
        else {
            break;
        }
    }
    return { args, requiredCount, restArg };
};
const isInvocable = (target) => ((typeof target === 'function')
    || (Utils.isDict(target) && ('__invoke' in target)));
const invoke = (target, rawArgs, referrer, context, thisArg) => {
    context.stack.push(Utils.formatFrameString(referrer, context));
    if (typeof target === 'function') {
        const returnValue = target(rawArgs, referrer, context, thisArg);
        context.stack.pop();
        return returnValue;
    }
    else if (Utils.isDict(target) && ('__invoke' in target)) {
        if (typeof target.__invoke !== 'function') {
            Utils.raise(TypeError, 'expect `__invoke` to be a function', referrer, context);
        }
        const returnValue = target.__invoke(rawArgs, referrer, context, thisArg);
        context.stack.pop();
        return returnValue;
    }
    else {
        Utils.raise(TypeError, 'expect a function or dict with proper `__invoke`', referrer, context);
        return null; // for type checking
    }
};

/**
 * Evalute an list.
 */
const evalList = (nodes, context, begin = 0, end = nodes.length) => {
    const result = [];
    if (begin >= end) {
        return result;
    }
    let left = begin;
    for (let right = 0; right <= end; right++) {
        if (right < end) {
            const node = nodes[right];
            if (node.type !== 'symbol' || node.value !== ',') {
                continue;
            }
        }
        const firstNode = nodes[left];
        if (firstNode.type === 'symbol' && firstNode.value === '...') {
            const values = evalExpression(nodes, context, left + 1, right);
            if (!Array.isArray(values)) {
                Utils.raise(TypeError, 'expect an array following', firstNode, context);
            }
            for (let i = 0; i < values.length; i++) {
                result.push(values[i]);
            }
        }
        else {
            result.push(evalExpression(nodes, context, left, right));
        }
        if (right < end - 1) {
            left = right + 1;
        }
        else {
            break;
        }
    }
    return result;
};

const createFunctionHandler = (minArgCount, maxArgCount, callback) => ((rawArgs, referrer, context, thisArg) => {
    const args = evalList(rawArgs, context);
    if (args.length < minArgCount) {
        Utils.raise(TypeError, 'too few arguments', referrer, context);
    }
    else if (args.length > maxArgCount) {
        Utils.raise(TypeError, 'too many arguments', referrer, context);
    }
    return callback(args, referrer, context, thisArg);
});

/**
 * Create an expression from source code.
 */
const createInlineExpression = (buffer, index, context) => {
    const argSpan = buffer[index - 1];
    const bodySpan = buffer[index + 1];
    if (index + 1 >= buffer.length
        || argSpan.type !== 'span'
        || argSpan.start !== '('
        || bodySpan.type !== 'span'
        || bodySpan.start !== '(') {
        Utils.raise(SyntaxError, 'invalid expression declaration', buffer[index], context);
    }
    const argList = parseArgList(argSpan.body, context);
    const body = bodySpan.body;
    const compiledBody = compileNodes(body, context);
    return createFunctionHandler(argList.requiredCount, Infinity, (args, referrer, _context, thisArg) => {
        const scopeStore = Utils.createDict(context.store);
        scopeStore._ = null;
        scopeStore.this = thisArg;
        scopeStore.arguments = args;
        const scopeContext = Utils.extendContext(context, scopeStore);
        const argDefinitions = argList.args;
        for (let i = 0; i < argDefinitions.length; i++) {
            const argName = argDefinitions[i].name;
            if (i < args.length) {
                scopeStore[argName] = args[i];
            }
            else {
                const defaultValue = argDefinitions[i].default;
                if (defaultValue !== null) {
                    scopeStore[argName] = evalCompiledExpression(defaultValue, scopeContext, true, true);
                }
                else {
                    scopeStore[argName] = null;
                }
            }
        }
        if (argList.restArg !== null) {
            scopeStore[argList.restArg] = args.slice(argDefinitions.length);
        }
        // rewrite
        scopeStore._ = null;
        scopeStore.this = thisArg;
        scopeStore.arguments = args;
        return evalCompiledNodes(compiledBody, scopeContext, true, true);
    });
};
const inlineExpressionHandler = (buffer, index, context) => {
    const func = createInlineExpression(buffer, index, context);
    const valueNode = Utils.createValueNode(func, buffer[index]);
    Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
};

/**
 * Create a function from source code.
 */
const createInlineFunction = (buffer, index, context) => {
    const argSpan = buffer[index + 1];
    const bodySpan = buffer[index + 2];
    if (index + 2 >= buffer.length
        || argSpan.type !== 'span'
        || argSpan.start !== '('
        || bodySpan.type !== 'span'
        || bodySpan.start !== '{') {
        Utils.raise(SyntaxError, 'invalid function declaration', buffer[index], context);
    }
    const argList = parseArgList(argSpan.body, context);
    const body = bodySpan.body;
    const compiledBody = compileNodes(body, context);
    return createFunctionHandler(argList.requiredCount, Infinity, (args, referrer, _context, thisArg) => {
        const scopeStore = Utils.createDict(context.store);
        scopeStore._ = null;
        scopeStore.this = thisArg;
        scopeStore.arguments = args;
        const scopeContext = Utils.extendContext(context, scopeStore);
        const argDefinitions = argList.args;
        for (let i = 0; i < argDefinitions.length; i++) {
            const argName = argDefinitions[i].name;
            if (i < args.length) {
                scopeStore[argName] = args[i];
            }
            else {
                const defaultValue = argDefinitions[i].default;
                if (defaultValue !== null) {
                    scopeStore[argName] = evalCompiledExpression(defaultValue, scopeContext, true, true);
                }
                else {
                    scopeStore[argName] = null;
                }
            }
        }
        if (argList.restArg !== null) {
            scopeStore[argList.restArg] = args.slice(argDefinitions.length);
        }
        // rewrite
        scopeStore._ = null;
        scopeStore.this = thisArg;
        scopeStore.arguments = args;
        const RETURN_FLAG = Symbol('hxs-return-flag');
        let returnValue = null;
        scopeStore.return = createFunctionHandler(0, 1, args => {
            if (args.length) {
                returnValue = args[0];
            }
            throw RETURN_FLAG;
        });
        scopeStore.forward = createFunctionHandler(1, 1, (args, _referrer) => {
            const names = args[0];
            if (!Array.isArray(names)) {
                Utils.raise(TypeError, 'expect an array of strings', _referrer, context);
            }
            for (let i = 0; i < names.length; i++) {
                const name = names[i];
                if (typeof name !== 'string') {
                    Utils.raise(TypeError, 'expect an array of strings', _referrer, context);
                }
                if (name in scopeStore) {
                    context.store[name] = scopeStore[name];
                }
                else {
                    Utils.raise(ReferenceError, `${Utils.toDisplay(name)} is not defined`, _referrer, context);
                }
            }
            return null;
        });
        try {
            evalCompiledNodes(compiledBody, scopeContext, true, true);
        }
        catch (err) {
            if (err !== RETURN_FLAG) {
                throw err;
            }
        }
        return returnValue;
    });
};
const inlineFunctionHandler = (buffer, index, context) => {
    const secondNode = buffer[index + 1];
    if (secondNode.type === 'word') { // named
        const name = secondNode.value;
        const func = createInlineFunction(buffer, index + 1, context);
        context.store[name] = func;
        const valueNode = Utils.createValueNode(func, buffer[index]);
        Utils.replaceBuffer(buffer, index, 4, valueNode);
    }
    else { // anonymous
        const func = createInlineFunction(buffer, index, context);
        const valueNode = Utils.createValueNode(func, buffer[index]);
        Utils.replaceBuffer(buffer, index, 3, valueNode);
    }
};

const getKeysOf = (dict, referrer, context) => {
    const keys = isInvocable(dict.__keys)
        ? invoke(dict.__keys, [], referrer, context, null)
        : Object.keys(dict);
    if (!Array.isArray(keys)) {
        Utils.raise(TypeError, '`__keys` should return an array', referrer, context);
    }
    return keys;
};
const hasProperty = (dict, key, keyReferrer, operationReferrer, context) => {
    if (isInvocable(dict.__has)) {
        return invoke(dict.__has, [Utils.createValueNode(key, keyReferrer)], operationReferrer, context, null);
    }
    else {
        if (typeof key !== 'string') {
            Utils.raise(TypeError, 'expect a string as key', keyReferrer, context);
        }
        const keys = getKeysOf(dict, operationReferrer, context);
        return keys.includes(key);
    }
};
const removeProperty = (target, key, keyReferrer, operationReferrer, context) => {
    if (Array.isArray(target)) {
        if (typeof key !== 'number') {
            Utils.raise(TypeError, 'expect a number as index', keyReferrer, context);
        }
        const normalizedIndex = Utils.normalizeIndex(key, target.length, keyReferrer, context);
        Utils.removeElements(target, normalizedIndex, 1);
    }
    else if (Utils.isDict(target)) {
        if (isInvocable(target.__remove)) {
            invoke(target.__remove, [Utils.createValueNode(key, keyReferrer)], operationReferrer, context, null);
        }
        else {
            if (typeof key !== 'string') {
                Utils.raise(TypeError, 'expect a string as key', keyReferrer, context);
            }
            if (key in target) { // Don't use `hasProperty` here!
                delete target[key];
            }
        }
    }
    else {
        Utils.raise(TypeError, 'invalid property removal', operationReferrer, context);
    }
};
const getProperty = (target, key, keyReferrer, operationReferrer, context) => {
    if (Array.isArray(target) || typeof target === 'string') {
        if (typeof key !== 'number') {
            Utils.raise(TypeError, 'expect a number as index', keyReferrer, context);
        }
        const normalizedIndex = Utils.normalizeIndex(key, target.length, keyReferrer, context);
        return target[normalizedIndex];
    }
    else if (Utils.isDict(target)) {
        if (isInvocable(target.__get)) {
            return invoke(target.__get, [Utils.createValueNode(key, keyReferrer)], operationReferrer, context, null);
        }
        else {
            if (typeof key !== 'string') {
                Utils.raise(TypeError, 'expect a string as key', keyReferrer, context);
            }
            if (key in target) { // Don't use `hasProperty` here!
                return target[key];
            }
            else {
                return null;
            }
        }
    }
    else {
        Utils.raise(TypeError, 'invalid property access', operationReferrer, context);
        return null; // for type checking
    }
};
const setProperty = (target, key, value, keyReferrer, operationReferrer, context) => {
    if (Array.isArray(target)) {
        if (typeof key !== 'number') {
            Utils.raise(TypeError, 'expect a number as index', keyReferrer, context);
        }
        const normalizedIndex = Utils.normalizeIndex(key, target.length, keyReferrer, context);
        target[normalizedIndex] = value;
    }
    else if (Utils.isDict(target)) {
        if (isInvocable(target.__set)) {
            invoke(target.__set, Utils.createRawArray([key, value], operationReferrer), operationReferrer, context, target);
        }
        else {
            if (typeof key !== 'string') {
                Utils.raise(TypeError, 'expect a string as key', keyReferrer, context);
            }
            target[key] = value;
        }
    }
    else {
        Utils.raise(TypeError, 'invalid property assignment', operationReferrer, context);
    }
};
const createClass = (description, referrer, context) => {
    const constructor = Utils.createDict({
        __invoke(_rawArgs, _referrer, _context, _thisArg) {
            let object = _thisArg;
            if (!Utils.isDict(object)) { // no prototype
                if (object !== null) {
                    Utils.raise(TypeError, 'expect only null or dict as this arg for class constructor', referrer, context);
                }
                object = Object.create(null);
                object[CONSTRUCTOR_SYMBOL] = constructor;
            }
            for (const key in description) {
                if (key in object) {
                    continue;
                }
                const value = description[key];
                if (typeof value === 'function') {
                    object[key] = (rawArgs, referrer, context) => (value(rawArgs, referrer, context, object));
                }
                else {
                    object[key] = value;
                }
            }
            if (description.__init) {
                invoke(description.__init, _rawArgs, _referrer, _context, object);
            }
            return object;
        },
    });
    return constructor;
};
const getConstructorOf = (dict) => ((CONSTRUCTOR_SYMBOL in dict)
    ? dict[CONSTRUCTOR_SYMBOL]
    : null);
const isInstanceOf = (constructor, dict) => {
    let dictConstructor = getConstructorOf(dict);
    if (constructor === dictConstructor) {
        return true;
    }
    while (dictConstructor) {
        if (Utils.isDict(dictConstructor) && (BASE_SYMBOL in dictConstructor)) {
            dictConstructor = dictConstructor[BASE_SYMBOL];
            if (constructor === dictConstructor) {
                return true;
            }
        }
        else {
            break;
        }
    }
    return false;
};

/**
 * Execute and replace the given node of the buffer.
 */
const evalBufferNode = (buffer, index, referrer, context) => {
    if (index < 0 || index >= buffer.length) {
        Utils.raise(SyntaxError, 'invalid operation', referrer, context);
    }
    return evalNode(buffer[index], context);
};

const evalBinaryOperation = (a, b, magicName, typeA, typeB, defaultHandler, buffer, index, context) => {
    if (magicName) {
        if (Utils.isDict(a) && (magicName in a)) {
            const magicMethod = a[magicName];
            if (!isInvocable(magicMethod)) {
                Utils.raise(TypeError, `expect an invocable as ${magicName}`, buffer[index - 1], context);
            }
            return invoke(magicMethod, Utils.createRawArray([b, false], buffer[index]), buffer[index], context, null);
        }
        else if (Utils.isDict(b) && (magicName in b)) {
            const magicMethod = b[magicName];
            if (!isInvocable(magicMethod)) {
                Utils.raise(TypeError, `expect an invocable as ${magicName}`, buffer[index + 1], context);
            }
            return invoke(magicMethod, Utils.createRawArray([a, true], buffer[index]), buffer[index], context, null);
        }
    }
    if (typeA && typeof a !== typeA) {
        Utils.raise(TypeError, `expect a ${typeA}`, buffer[index - 1], context);
    }
    if (typeB && typeof b !== typeB) {
        Utils.raise(TypeError, `expect a ${typeB}`, buffer[index + 1], context);
    }
    return defaultHandler(a, b);
};
const evalUnaryOperation = (value, magicName, type, defaultHandler, buffer, index, context) => {
    if (magicName && Utils.isDict(value) && (magicName in value)) {
        const magicMethod = value[magicName];
        if (!isInvocable(magicMethod)) {
            Utils.raise(TypeError, `expect an invocable as ${magicName}`, buffer[index - 1], context);
        }
        return invoke(magicMethod, [Utils.createValueNode(value, buffer[index])], buffer[index], context, null);
    }
    else {
        if (type && typeof value !== type) {
            Utils.raise(TypeError, `expect a ${type}`, buffer[index - 1], context);
        }
        return defaultHandler(value);
    }
};
const createBinaryOperator = (magicName, typeA, typeB, defaultHandler) => ((buffer, index, context) => {
    const a = evalBufferNode(buffer, index - 1, buffer[index], context);
    const b = evalBufferNode(buffer, index + 1, buffer[index], context);
    const valueNode = Utils.createValueNode(evalBinaryOperation(a, b, magicName, typeA, typeB, defaultHandler, buffer, index, context), buffer[index]);
    Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
});
const createUnaryOperator = (magicName, type, defaultHandler) => ((buffer, index, context) => {
    const value = evalBufferNode(buffer, index + 1, buffer[index], context);
    const valueNode = Utils.createValueNode(evalUnaryOperation(value, magicName, type, defaultHandler, buffer, index, context), buffer[index]);
    Utils.replaceBuffer(buffer, index, 2, valueNode);
});
const createAdditionalAssignmentOperator = (magicName, typeA, typeB, defaultHandler) => ((buffer, index, context) => {
    const operationReferrer = buffer[index];
    if (index === 0) {
        Utils.raise(SyntaxError, 'expect a variable preceding', operationReferrer, context);
    }
    const precedingNode = buffer[index - 1];
    let target = null;
    let key = null;
    if (precedingNode.type === 'word') { // a = b or a.b = c
        if (index === 1) { // a = b
            const name = precedingNode.value;
            if (!(name in context.store)) {
                Utils.raise(SyntaxError, `${Utils.toDisplay(name)} is not defined`, precedingNode, context);
            }
            target = context.store;
            key = precedingNode.value;
        }
        else { // a.b = c
            if (index === 2
                || buffer[index - 2].type !== 'symbol'
                || buffer[index - 2].value !== '.') {
                Utils.raise(SyntaxError, 'invalid assignment', operationReferrer, context);
            }
            target = evalExpression(buffer, context, 0, index - 2);
            if (!Utils.isDict(target)) {
                Utils.raise(TypeError, 'expect a dict as assignment target', buffer[0], context);
            }
            key = precedingNode.value;
        }
    }
    else if (precedingNode.type === 'span'
        && precedingNode.start === '[') { // a[b] = c
        target = evalExpression(buffer, context, 0, index - 1);
        key = evalExpression(precedingNode.body, context);
    }
    else {
        Utils.raise(SyntaxError, 'invalid assignment', operationReferrer, context);
    }
    const a = getProperty(target, key, precedingNode, operationReferrer, context);
    const b = evalExpression(buffer, context, index + 1);
    const value = evalBinaryOperation(a, b, magicName, typeA, typeB, defaultHandler, buffer, index, context);
    setProperty(target, key, value, precedingNode, operationReferrer, context);
    const valueNode = Utils.createValueNode(value, precedingNode);
    Utils.replaceBuffer(buffer, 0, buffer.length, valueNode);
});

const assignmentOperators = [{
        symbol: '=',
        priority: -Infinity,
        ltr: true,
        handler(buffer, index, context) {
            const operationReferrer = buffer[index];
            if (index === 0) {
                Utils.raise(SyntaxError, 'expect a variable preceding', operationReferrer, context);
            }
            const precedingNode = buffer[index - 1];
            const value = evalExpression(buffer, context, index + 1);
            if (precedingNode.type === 'word') { // a = b or a.b = c
                if (index === 1) { // a = b
                    context.store[precedingNode.value] = value;
                }
                else { // a.b = c
                    if (index === 2
                        || buffer[index - 2].type !== 'symbol'
                        || buffer[index - 2].value !== '.') {
                        Utils.raise(SyntaxError, 'invalid assignment', operationReferrer, context);
                    }
                    const target = evalExpression(buffer, context, 0, index - 2);
                    if (!Utils.isDict(target)) {
                        Utils.raise(TypeError, 'expect a dict as assignment target', buffer[0], context);
                    }
                    setProperty(target, precedingNode.value, value, buffer[0], operationReferrer, context);
                }
            }
            else if (precedingNode.type === 'span'
                && precedingNode.start === '[') { // a[b] = c
                const target = evalExpression(buffer, context, 0, index - 1);
                const key = evalExpression(precedingNode.body, context);
                setProperty(target, key, value, buffer[0], operationReferrer, context);
            }
            else {
                Utils.raise(SyntaxError, 'invalid assignment', operationReferrer, context);
            }
            const valueNode = Utils.createValueNode(value, operationReferrer);
            Utils.replaceBuffer(buffer, 0, buffer.length, valueNode);
        },
    }, {
        symbol: '+=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__plus', 'number', 'number', (a, b) => (a + b)),
    }, {
        symbol: '-=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__minus', 'number', 'number', (a, b) => (a - b)),
    }, {
        symbol: '*=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__multiply', 'number', 'number', (a, b) => (a * b)),
    }, {
        symbol: '/=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__divide', 'number', 'number', (a, b) => (a / b)),
    }, {
        symbol: '%=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__mod', 'number', 'number', (a, b) => (a % b)),
    }, {
        symbol: '**=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__pow', 'number', 'number', (a, b) => (a ** b)),
    }, {
        symbol: '&=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__bitAnd', 'number', 'number', (a, b) => (a & b)),
    }, {
        symbol: '^=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__bitXor', 'number', 'number', (a, b) => (a ^ b)),
    }, {
        symbol: '|=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__bitOr', 'number', 'number', (a, b) => (a | b)),
    }, {
        symbol: '<<=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__leftShift', 'number', 'number', (a, b) => (a << b)),
    }, {
        symbol: '>>=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__rightShift', 'number', 'number', (a, b) => (a >> b)),
    }, {
        symbol: '>>>=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__unsignedRightShift', 'number', 'number', (a, b) => (a >>> b)),
    }, {
        symbol: '&&=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__and', 'boolean', 'boolean', (a, b) => (a && b)),
    }, {
        symbol: '||=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__or', 'boolean', 'boolean', (a, b) => (a || b)),
    }, {
        symbol: '??=',
        priority: -Infinity,
        ltr: true,
        handler: createAdditionalAssignmentOperator('__nullOr', null, null, (a, b) => (a ?? b)),
    }];

const bitOperators = [{
        symbol: '~',
        priority: 2,
        ltr: false,
        handler: createUnaryOperator('__bitNot', 'number', (x) => (~x)),
    }, {
        symbol: '<<',
        priority: 5,
        ltr: true,
        handler: createBinaryOperator('__leftShift', 'number', 'number', (a, b) => (a << b)),
    }, {
        symbol: '>>',
        priority: 5,
        ltr: true,
        handler: createBinaryOperator('__rightShift', 'number', 'number', (a, b) => (a >> b)),
    }, {
        symbol: '>>>',
        priority: 5,
        ltr: true,
        handler: createBinaryOperator('__unsignedRightShift', 'number', 'number', (a, b) => (a >>> b)),
    }, {
        symbol: '&',
        priority: 8,
        ltr: true,
        handler: createBinaryOperator('__bitAnd', 'number', 'number', (a, b) => (a & b)),
    }, {
        symbol: '^',
        priority: 9,
        ltr: true,
        handler: createBinaryOperator('__bitXor', 'number', 'number', (a, b) => (a ^ b)),
    }, {
        symbol: '|',
        priority: 10,
        ltr: true,
        handler: createBinaryOperator('__bitOr', 'number', 'number', (a, b) => (a | b)),
    }];

const booleanOperators = [{
        symbol: '!',
        priority: 2,
        ltr: false,
        handler: createUnaryOperator('__not', 'boolean', (x) => (!x)),
    }, {
        symbol: '&&',
        priority: 11,
        ltr: true,
        handler: createBinaryOperator('__and', 'boolean', 'boolean', (a, b) => (a && b)),
    }, {
        symbol: '||',
        priority: 12,
        ltr: true,
        handler: createBinaryOperator('__or', 'boolean', 'boolean', (a, b) => (a || b)),
    }];

const braceHandler = (buffer, index, context) => {
    if (index === 0 || buffer[index - 1].type === 'symbol') { // array creation
        const value = evalList(buffer[index].body, context);
        buffer[index] = Utils.createValueNode(value, buffer[index]);
    }
    else { // key access
        const target = evalBufferNode(buffer, index - 1, buffer[index], context);
        const key = evalExpression(buffer[index].body, context);
        const value = getProperty(target, key, buffer[index - 1], buffer[index], context);
        const valueNode = Utils.createValueNode(value, buffer[index]);
        Utils.replaceBuffer(buffer, index - 1, 2, valueNode);
    }
};

const bracketHandler = (buffer, index, context) => {
    if (index === 0 || buffer[index - 1].type === 'symbol') { // dict creation
        const result = Object.create(null);
        const nodes = buffer[index].body;
        if (nodes.length === 0) {
            buffer[index] = Utils.createValueNode(result, buffer[index]);
            return;
        }
        let left = 0;
        for (let right = 0; right <= nodes.length; right++) {
            if (right < nodes.length) {
                const node = nodes[right];
                if (node.type !== 'symbol' || node.value !== ',') {
                    continue;
                }
            }
            const firstNode = nodes[left];
            if (firstNode.type === 'symbol' && firstNode.value === '...') { // expand
                const props = evalExpression(nodes, context, left + 1, right);
                if (!Utils.isDict(props)) {
                    Utils.raise(TypeError, 'expect a dict to expand', firstNode, context);
                }
                Object.assign(result, props);
            }
            else { // insert
                let j;
                for (j = left; j < right; j++) {
                    if (nodes[j].type === 'symbol'
                        && nodes[j].value === '->') {
                        break;
                    }
                }
                if (j === right) { // shortcut
                    if (firstNode.type === 'symbol' && firstNode.value === '@') { // function
                        if (right < left + 3) {
                            Utils.raise(SyntaxError, 'invalid function declaration', firstNode, context);
                        }
                        const nameNode = nodes[left + 1];
                        if (nameNode.type !== 'word') {
                            Utils.raise(SyntaxError, 'invalid function declaration', firstNode, context);
                        }
                        const name = nameNode.value;
                        const func = createInlineFunction(nodes, left + 1, context);
                        result[name] = func;
                    }
                    else { // variable
                        const name = evalExpression(nodes, context, left, right);
                        if (typeof name !== 'string') {
                            Utils.raise(TypeError, 'expect a string as key', firstNode, context);
                        }
                        if (!(name in context.store)) {
                            Utils.raise(ReferenceError, `${Utils.toDisplay(name)} is not defined`, firstNode, context);
                        }
                        result[name] = context.store[name];
                    }
                }
                else { // key -> value
                    const name = evalExpression(nodes, context, left, j);
                    if (typeof name !== 'string') {
                        Utils.raise(TypeError, 'expect a string as key', firstNode, context);
                    }
                    const value = evalExpression(nodes, context, j + 1, right);
                    result[name] = value;
                }
            }
            if (right < nodes.length - 1) {
                left = right + 1;
            }
            else {
                break;
            }
        }
        buffer[index] = Utils.createValueNode(result, buffer[index]);
    }
    else { // callback invocation
        const handler = evalNode(buffer[index - 1], context);
        if (!isInvocable(handler)) {
            Utils.raise(TypeError, 'expect an invocable', buffer[index], context);
        }
        const compiledBody = compileNodes(buffer[index].body, context);
        const callback = (args, referrer, _context) => {
            return evalCompiledNodes(compiledBody, _context, true, true);
        };
        const returnValue = invoke(handler, [Utils.createValueNode(callback, buffer[index])], buffer[index], context, null);
        const valueNode = Utils.createValueNode(returnValue, buffer[index]);
        Utils.replaceBuffer(buffer, index - 1, 2, valueNode);
    }
};

const compareOperators = [{
        symbol: '<',
        priority: 6,
        ltr: true,
        handler: createBinaryOperator('__lt', 'number', 'number', (a, b) => (a < b)),
    }, {
        symbol: '>',
        priority: 6,
        ltr: true,
        handler: createBinaryOperator('__gt', 'number', 'number', (a, b) => (a > b)),
    }, {
        symbol: '<=',
        priority: 6,
        ltr: true,
        handler: createBinaryOperator('__lte', 'number', 'number', (a, b) => (a <= b)),
    }, {
        symbol: '>=',
        priority: 6,
        ltr: true,
        handler: createBinaryOperator('__gte', 'number', 'number', (a, b) => (a >= b)),
    }, {
        symbol: '==',
        priority: 7,
        ltr: true,
        handler: createBinaryOperator('__equal', null, null, (a, b) => (a === b)),
    }, {
        symbol: '!=',
        priority: 7,
        ltr: true,
        handler: createBinaryOperator('__notEqual', null, null, (a, b) => (a !== b)),
    }, {
        symbol: '===',
        priority: 7,
        ltr: true,
        handler(buffer, index, context) {
            const a = evalBufferNode(buffer, index - 1, buffer[index], context);
            const b = evalBufferNode(buffer, index + 1, buffer[index], context);
            const valueNode = Utils.createValueNode(a === b, buffer[index]);
            Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
        },
    }, {
        symbol: '!==',
        priority: 7,
        ltr: true,
        handler(buffer, index, context) {
            const a = evalBufferNode(buffer, index - 1, buffer[index], context);
            const b = evalBufferNode(buffer, index + 1, buffer[index], context);
            const valueNode = Utils.createValueNode(a !== b, buffer[index]);
            Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
        },
    }];

const mathOperators = [{
        symbol: '**',
        priority: 2,
        ltr: false,
        handler: createBinaryOperator('__pow', 'number', 'number', Math.pow),
    }, {
        symbol: '*',
        priority: 3,
        ltr: true,
        handler: createBinaryOperator('__multiply', 'number', 'number', (a, b) => (a * b)),
    }, {
        symbol: '/',
        priority: 3,
        ltr: true,
        handler: createBinaryOperator('__divide', 'number', 'number', (a, b) => (a / b)),
    }, {
        symbol: '%',
        priority: 3,
        ltr: true,
        handler: createBinaryOperator('__mod', 'number', 'number', (a, b) => (a % b)),
    }, {
        symbol: '+',
        priority: 4,
        ltr: true,
        handler: createBinaryOperator('__plus', 'number', 'number', (a, b) => (a + b)),
    }, {
        symbol: '-',
        priority: 4,
        ltr: true,
        handler(buffer, index, context) {
            const b = evalBufferNode(buffer, index + 1, buffer[index], context);
            if (index === 0 || buffer[index - 1].type === 'symbol') { // unary
                const result = evalUnaryOperation(b, '__opposite', 'number', (x) => (-x), buffer, index, context);
                const valueNode = Utils.createValueNode(result, buffer[index]);
                Utils.replaceBuffer(buffer, index, 2, valueNode);
            }
            else { // binary
                const a = evalBufferNode(buffer, index - 1, buffer[index], context);
                const result = evalBinaryOperation(a, b, '__minus', 'number', 'number', (a, b) => (a - b), buffer, index, context);
                const valueNode = Utils.createValueNode(result, buffer[index]);
                Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
            }
        },
    }];

const builtinArray = Utils.injectHelp('A dict providing methods for array manipulations.', Utils.createDict({
    __invoke: Utils.injectHelp('Array.__invoke(data...)', createFunctionHandler(0, Infinity, (args) => (args.slice()))),
    create: Utils.injectHelp('Array.create(size = 0, init = null)', createFunctionHandler(0, 2, (args, referrer, context) => {
        if (args.length === 0) {
            return [];
        }
        const count = args[0];
        if (typeof count !== 'number') {
            Utils.raise(TypeError, 'expect a number as array size', referrer, context);
        }
        if (count < 0 || !Number.isFinite(count)) {
            Utils.raise(RangeError, 'invalid array size', referrer, context);
        }
        const init = args.length > 1 ? args[1] : null;
        return Array.from({ length: count }, _ => init);
    })),
    clone: Utils.injectHelp('Array.clone(array)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to clone', referrer, context);
        }
        return array.slice();
    })),
    sizeOf: Utils.injectHelp('Array.sizeOf(array)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array', referrer, context);
        }
        return array.length;
    })),
    push: Utils.injectHelp('Array.push(array, data...)', createFunctionHandler(2, Infinity, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to modify', referrer, context);
        }
        for (let i = 1; i < args.length; i++) {
            array.push(args[i]);
        }
        return null;
    })),
    unshift: Utils.injectHelp('Array.unshift(array, data...)', createFunctionHandler(2, Infinity, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to modify', referrer, context);
        }
        for (let i = 1; i < args.length; i++) {
            array.unshift(args[i]);
        }
        return null;
    })),
    pop: Utils.injectHelp('Array.pop(array)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to modify', referrer, context);
        }
        return (array.length
            ? array.pop()
            : null);
    })),
    shift: Utils.injectHelp('Array.shift(array)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to modify', referrer, context);
        }
        return (array.length
            ? array.shift()
            : null);
    })),
    slice: Utils.injectHelp('Array.slice(array, start = 0, end = Array.sizeOf(array))', createFunctionHandler(1, 3, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array as source', referrer, context);
        }
        if (args.length > 1) {
            if (typeof args[1] !== 'number') {
                Utils.raise(TypeError, 'expect a number as begin index', referrer, context);
            }
            if (args.length > 2) {
                if (typeof args[2] !== 'number') {
                    Utils.raise(TypeError, 'expect a number as end index', referrer, context);
                }
            }
        }
        return array.slice(args[1], args[2]);
    })),
    insert: Utils.injectHelp('Array.insert(array, index, data...)', createFunctionHandler(3, Infinity, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to operate on', referrer, context);
        }
        const index = args[1];
        if (typeof index !== 'number') {
            Utils.raise(TypeError, 'expect a number as start index', referrer, context);
        }
        const arraySize = array.length;
        const normalizedIndex = Utils.normalizeIndex(index, array.length, referrer, context, true);
        const insertCount = args.length - 2;
        array.length += insertCount;
        // move subsequent elements
        for (let i = arraySize - 1; i >= normalizedIndex; i--) {
            array[i + insertCount] = array[i];
        }
        // insert new elements
        for (let i = 2; i < args.length; i++) {
            array[normalizedIndex + i - 2] = args[i];
        }
        return null;
    })),
    remove: Utils.injectHelp('Array.remove(array, index, count = 1)', createFunctionHandler(2, 3, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }
        const index = args[1];
        if (typeof index !== 'number') {
            Utils.raise(TypeError, 'expect a number as start index', referrer, context);
        }
        const arraySize = array.length;
        const normalizedIndex = Utils.normalizeIndex(index, array.length, referrer, context);
        if (args.length === 3) {
            if (typeof args[2] !== 'number') {
                Utils.raise(TypeError, 'expect a number as removing count', referrer, context);
            }
            if (args[2] < 0 || args[2] !== args[2]) {
                Utils.raise(RangeError, 'invalid removing count', referrer, context);
            }
        }
        const count = args.length === 3 ? args[2] : 1;
        if (count === 0) {
            return null;
        }
        if (normalizedIndex + count >= arraySize) {
            array.length = normalizedIndex;
        }
        else {
            // move trailing elements
            for (let i = normalizedIndex; i < arraySize - count; i++) {
                array[i] = array[i + count];
            }
            array.length -= count;
        }
        return null;
    })),
    clear: Utils.injectHelp('Array.clear(array)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to operate on', referrer, context);
        }
        array.length = 0;
        return null;
    })),
    flat: Utils.injectHelp('Array.flat(arrays, depth = 1)', createFunctionHandler(1, 2, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to flat', referrer, context);
        }
        const depth = (args.length === 2) ? args[1] : 1;
        if (typeof depth !== 'number') {
            Utils.raise(TypeError, 'expect a number as depth', referrer, context);
        }
        if (depth <= 0 || depth !== depth) {
            Utils.raise(RangeError, 'invalid depth', referrer, context);
        }
        return array.flat(depth);
    })),
    unpack: Utils.injectHelp('Array.unpack(array, names, loose = false)', createFunctionHandler(2, 3, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to unpack', referrer, context);
        }
        const names = args[1];
        if (!Array.isArray(names)) {
            Utils.raise(TypeError, 'expect an array of strings as variable names', referrer, context);
        }
        const nameCount = names.length;
        const arraySize = array.length;
        for (let i = 0; i < nameCount; i++) {
            if (typeof names[i] !== 'string') {
                Utils.raise(TypeError, 'expect strings as variable names', referrer, context);
            }
        }
        const loose = args.length === 3 && args[2];
        if (typeof loose !== 'boolean') {
            Utils.raise(TypeError, 'expect a boolean as loose option', referrer, context);
        }
        const { store } = context;
        if (loose) {
            if (nameCount > arraySize) {
                for (let i = 0; i < arraySize; i++) {
                    store[names[i]] = array[i];
                }
                for (let i = arraySize; i < nameCount; i++) {
                    store[names[i]] = null;
                }
            }
            else {
                for (let i = 0; i < nameCount; i++) {
                    store[names[i]] = array[i];
                }
            }
        }
        else {
            if (nameCount > arraySize) {
                Utils.raise(RangeError, 'not enough values in the given array', referrer, context);
            }
            for (let i = 0; i < nameCount; i++) {
                store[names[i]] = array[i];
            }
        }
        return null;
    })),
    indexOf: Utils.injectHelp('Array.indexOf(array, value)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to search in', referrer, context);
        }
        return array.indexOf(args[1]);
    })),
    lastIndexOf: Utils.injectHelp('Array.lastIndexOf(array, value)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to search in', referrer, context);
        }
        return array.lastIndexOf(args[1]);
    })),
    includes: Utils.injectHelp('Array.includes(array, value)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array to check', referrer, context);
        }
        return array.includes(args[1]);
    })),
    sort: Utils.injectHelp('Array.sort(array, compareFn?)', createFunctionHandler(1, 2, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }
        const compareFn = args[1];
        let _compareFn;
        if (args.length > 1 && !isInvocable(compareFn)) {
            Utils.raise(TypeError, 'expect an invocable as compare function', referrer, context);
        }
        if (compareFn) {
            _compareFn = (a, b) => invoke(compareFn, Utils.createRawArray([a, b], referrer), referrer, context, null);
        }
        return array.sort(_compareFn);
    })),
    forEach: Utils.injectHelp('Array.forEach(array, callback)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }
        const callback = args[1];
        if (!isInvocable(callback)) {
            Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
        }
        for (let i = 0; i < array.length; i++) {
            invoke(callback, Utils.createRawArray([array[i], i, array], referrer), referrer, context, null);
        }
        return null;
    })),
    map: Utils.injectHelp('Array.map(array, callback)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }
        const callback = args[1];
        if (!isInvocable(callback)) {
            Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
        }
        return array.map((element, index) => (invoke(callback, Utils.createRawArray([element, index, array], referrer), referrer, context, null)));
    })),
    filter: Utils.injectHelp('Array.filter(array, callback)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const array = args[0];
        if (!Array.isArray(array)) {
            Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }
        const callback = args[1];
        if (!isInvocable(callback)) {
            Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
        }
        return array.filter((element, index) => (invoke(callback, Utils.createRawArray([element, index, array], referrer), referrer, context, null)));
    })),
}));

const builtinDict = Utils.injectHelp('A dict providing methods for dict manipulations.', Utils.createDict({
    __invoke: Utils.injectHelp('Dict.__invoke(entries?)', createFunctionHandler(0, 1, (args, referrer, context) => {
        const dict = Object.create(null);
        if (args.length === 0) {
            return dict;
        }
        const entries = args[0];
        if (!Array.isArray(entries)) {
            Utils.raise(TypeError, `expect an array of arrays as entries`, referrer, context);
        }
        for (let i = 0; i < entries.length; i++) {
            const entry = entries[i];
            if (!Array.isArray(entry)
                || entry.length !== 2
                || typeof entry[0] !== 'string') {
                Utils.raise(TypeError, `invalid dict entry`, referrer, context);
            }
            dict[entry[0]] = entry[1];
        }
        return dict;
    })),
    clone: Utils.injectHelp('Dict.clone(dict)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const dict = args[0];
        if (!Utils.isDict(dict)) {
            Utils.raise(TypeError, `expect a dict to clone`, referrer, context);
        }
        return Utils.createDict(dict);
    })),
    has: Utils.injectHelp('Dict.has(dict, key)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const dict = args[0];
        if (!Utils.isDict(dict)) {
            Utils.raise(TypeError, `expect a dict to check`, referrer, context);
        }
        const key = args[1];
        if (typeof key !== 'string') {
            Utils.raise(TypeError, 'expect a string as key', referrer, context);
        }
        return (key in dict);
    })),
    remove: Utils.injectHelp('Dict.remove(dict, key)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const dict = args[0];
        if (!Utils.isDict(dict)) {
            Utils.raise(TypeError, `expect a dict`, referrer, context);
        }
        const key = args[1];
        if (typeof key !== 'string') {
            Utils.raise(TypeError, 'expect a string as key', referrer, context);
        }
        if (key in dict) { // Don't use `hasProperty` here!
            delete dict[key];
        }
        return null;
    })),
    get: Utils.injectHelp('Dict.get(dict, key)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const dict = args[0];
        if (!Utils.isDict(dict)) {
            Utils.raise(TypeError, `expect a dict`, referrer, context);
        }
        const key = args[1];
        if (typeof key !== 'string') {
            Utils.raise(TypeError, 'expect a string as key', referrer, context);
        }
        if (key in dict) { // Don't use `hasProperty` here!
            return dict[key];
        }
        else {
            return null;
        }
    })),
    set: Utils.injectHelp('Dict.set(dict, key, value)', createFunctionHandler(3, 3, (args, referrer, context) => {
        const dict = args[0];
        if (!Utils.isDict(dict)) {
            Utils.raise(TypeError, `expect a dict`, referrer, context);
        }
        const key = args[1];
        if (typeof key !== 'string') {
            Utils.raise(TypeError, 'expect a string as key', referrer, context);
        }
        dict[key] = args[2];
        return null;
    })),
    keys: Utils.injectHelp('Dict.keys(dict)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const dict = args[0];
        if (!Utils.isDict(dict)) {
            Utils.raise(TypeError, `expect a dict`, referrer, context);
        }
        return Object.keys(dict);
    })),
    entries: Utils.injectHelp('Dict.entries(dict)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const dict = args[0];
        if (!Utils.isDict(dict)) {
            Utils.raise(TypeError, `expect a dict`, referrer, context);
        }
        return Object.entries(dict);
    })),
    unpack: Utils.injectHelp('Dict.unpack(dict, names, loose = false)', createFunctionHandler(2, 3, (args, referrer, context) => {
        const dict = args[0];
        if (!Utils.isDict(dict)) {
            Utils.raise(TypeError, `expect a dict to unpack`, referrer, context);
        }
        const names = args[1];
        if (!Array.isArray(names)) {
            Utils.raise(TypeError, `expect an array of strings as variable names`, referrer, context);
        }
        for (let i = 0; i < names.length; i++) {
            if (typeof names[i] !== 'string') {
                Utils.raise(TypeError, `expect strings as variable names`, referrer, context);
            }
        }
        const loose = args.length === 3 && args[2];
        if (typeof loose !== 'boolean') {
            Utils.raise(TypeError, `expect a boolean as loose option`, referrer, context);
        }
        const { store } = context;
        if (loose) {
            for (let i = 0; i < names.length; i++) {
                if (hasProperty(dict, names[i], referrer, referrer, context)) {
                    store[names[i]] = getProperty(dict, names[i], referrer, referrer, context);
                }
                else {
                    store[names[i]] = null;
                }
            }
        }
        else {
            for (let i = 0; i < names.length; i++) {
                if (!hasProperty(dict, names[i], referrer, referrer, context)) {
                    Utils.raise(ReferenceError, `unknown key(${Utils.toDisplay(names[i])})`, referrer, context);
                }
            }
            for (let i = 0; i < names.length; i++) {
                store[names[i]] = getProperty(dict, names[i], referrer, referrer, context);
            }
        }
        return null;
    })),
    assign: Utils.injectHelp('Dict.assign(dict, props, allowOverwrite = true)', createFunctionHandler(2, 3, (args, referrer, context) => {
        const dict = args[0];
        if (!Utils.isDict(dict)) {
            Utils.raise(TypeError, `expect a dict to modify`, referrer, context);
        }
        const props = args[1];
        if (!Utils.isDict(props)) {
            Utils.raise(TypeError, `expect another dict as props`, referrer, context);
        }
        const allowOverwrite = args.length === 3 ? args[2] : true;
        if (typeof allowOverwrite !== 'boolean') {
            Utils.raise(TypeError, `expect a boolean as overwrite option`, referrer, context);
        }
        getKeysOf(props, referrer, context).forEach(key => {
            if (!allowOverwrite
                && hasProperty(dict, key, referrer, referrer, context)) {
                return;
            }
            const value = getProperty(props, key, referrer, referrer, context);
            setProperty(dict, key, value, referrer, referrer, context);
        });
        return dict;
    })),
    getConstructorOf: Utils.injectHelp('Dict.getConstructorOf(dict)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const dict = args[0];
        if (!Utils.isDict(dict)) {
            Utils.raise(TypeError, 'expect a dict', referrer, context);
        }
        return getConstructorOf(dict);
    })),
    isInstanceOf: Utils.injectHelp('Dict.isInstanceOf(dict, class)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const dict = args[0];
        if (!Utils.isDict(dict)) {
            Utils.raise(TypeError, 'expect a dict to check', referrer, context);
        }
        return isInstanceOf(args[1], dict);
    })),
}));

const builtinFunction = Utils.injectHelp('A dict providing methods for function manipulations.', Utils.createDict({
    isInvocable: Utils.injectHelp('Function.isInvocable(target)', createFunctionHandler(1, 1, (args, referrer, context) => (isInvocable(args[0])))),
    invoke: Utils.injectHelp('Function.invoke(invocable, args = null, thisArg = null)', createFunctionHandler(1, 3, (args, referrer, context) => {
        const fn = args[0];
        if (!isInvocable(fn)) {
            Utils.raise(TypeError, 'expect an invocable to invoke', referrer, context);
        }
        const fnArgs = args.length > 1
            ? args[1]
            : null;
        if (fnArgs !== null && !Array.isArray(fnArgs)) {
            Utils.raise(TypeError, 'expect an array or null as arguments', referrer, context);
        }
        const thisArg = args.length === 3 ? args[2] : null;
        return invoke(fn, fnArgs ? Utils.createRawArray(fnArgs, referrer) : [], referrer, context, thisArg);
    })),
    bind: Utils.injectHelp('Function.bind(invocable, thisArg)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const fn = args[0];
        if (!isInvocable(fn)) {
            Utils.raise(TypeError, 'expect an invocable to bind', referrer, context);
        }
        return (_args, _referrer, _context) => {
            return invoke(fn, _args, _referrer, _context, args[1]);
        };
    })),
}));

const builtinNumber = Utils.injectHelp('A dict providing methods for number manipulations.', Utils.createDict({
    toString: Utils.injectHelp('Number.toString(number, radix = 10)', createFunctionHandler(1, 2, (args, referrer, context) => {
        const number = args[0];
        if (typeof number !== 'number') {
            Utils.raise(TypeError, 'expect a number to convert', referrer, context);
        }
        if (args.length === 1) {
            return number.toString();
        }
        const radix = args[1];
        if (typeof radix !== 'number') {
            Utils.raise(TypeError, 'expect a number as radix', referrer, context);
        }
        if (!(radix >= 2 && radix <= 36)) {
            Utils.raise(RangeError, 'radix must be between 2 and 36', referrer, context);
        }
        return number.toString(radix);
    })),
    toFixed: Utils.injectHelp('Number.toFixed(number, fractionDigits = 0)', createFunctionHandler(1, 2, (args, referrer, context) => {
        const number = args[0];
        if (typeof number !== 'number') {
            Utils.raise(TypeError, 'expect a number to convert', referrer, context);
        }
        if (args.length === 1) {
            return number.toFixed();
        }
        const fractionDigits = args[1];
        if (typeof fractionDigits !== 'number') {
            Utils.raise(TypeError, 'expect a number as count of fraction digits', referrer, context);
        }
        if (!(fractionDigits >= 0 && fractionDigits <= 20)) {
            Utils.raise(RangeError, 'fractionDigits must be between 0 and 20', referrer, context);
        }
        return number.toFixed(fractionDigits);
    })),
    toExponential: Utils.injectHelp('Number.toExponential(number, fractionDigits = 0)', createFunctionHandler(1, 2, (args, referrer, context) => {
        const number = args[0];
        if (typeof number !== 'number') {
            Utils.raise(TypeError, 'expect a number to convert', referrer, context);
        }
        if (args.length === 1) {
            return number.toExponential();
        }
        const fractionDigits = args[1];
        if (typeof fractionDigits !== 'number') {
            Utils.raise(TypeError, 'expect a number as count of fraction digits', referrer, context);
        }
        if (!(fractionDigits >= 0 && fractionDigits <= 20)) {
            Utils.raise(RangeError, 'fractionDigits must be between 0 and 20', referrer, context);
        }
        return number.toExponential(fractionDigits);
    })),
    isFinite: Utils.injectHelp('Number.isFinite(number)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const number = args[0];
        if (typeof number !== 'number') {
            Utils.raise(TypeError, 'expect a number', referrer, context);
        }
        return Number.isFinite(number);
    })),
    isInteger: Utils.injectHelp('Number.isInteger(number)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const number = args[0];
        if (typeof number !== 'number') {
            Utils.raise(TypeError, 'expect a number', referrer, context);
        }
        return Number.isInteger(number);
    })),
    isNaN: Utils.injectHelp('Number.isNaN(number)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const number = args[0];
        if (typeof number !== 'number') {
            Utils.raise(TypeError, 'expect a number', referrer, context);
        }
        return Number.isNaN(number);
    })),
    parseInt: Utils.injectHelp('Number.parseInt(string, radix = 10)', createFunctionHandler(1, 2, (args, referrer, context) => {
        const string = args[0];
        if (typeof string !== 'string') {
            Utils.raise(TypeError, 'expect a string to convert', referrer, context);
        }
        const radix = args.length > 1 ? args[1] : 10;
        if (typeof radix !== 'number') {
            Utils.raise(TypeError, 'expect a number as radix', referrer, context);
        }
        if (!(radix >= 2 && radix <= 36)) {
            Utils.raise(RangeError, 'radix must be between 2 and 36', referrer, context);
        }
        const DIGITS = '0123456789abcdefghijklmnopqrstuvwxyz';
        let result = 0;
        for (let i = 0; i < string.length; i++) {
            const x = DIGITS.indexOf(string[string.length - i - 1]);
            if (x === -1 || x >= radix) {
                return NaN;
            }
            result += x * (radix ** i);
        }
        return result;
    })),
}));

const builtinString = Utils.injectHelp('A dict providing methods for string manipulations.', Utils.createDict({
    __invoke: Utils.injectHelp(`String.__invoke(parts...)`, createFunctionHandler(1, Infinity, (args, referrer, context) => {
        let result = '';
        for (let i = 0; i < args.length; i++) {
            result += Utils.toString(args[i]);
        }
        return result;
    })),
    sizeOf: Utils.injectHelp('String.sizeOf(string)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const string = args[0];
        if (typeof string !== 'string') {
            Utils.raise(TypeError, 'expect a string', referrer, context);
        }
        return string.length;
    })),
    toLowerCase: Utils.injectHelp(`String.toLowerCase(string)`, createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'string') {
            Utils.raise(TypeError, `expect a string`, referrer, context);
        }
        return args[0].toLowerCase();
    })),
    toUpperCase: Utils.injectHelp(`String.toUpperCase(string)`, createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'string') {
            Utils.raise(TypeError, `expect a string`, referrer, context);
        }
        return args[0].toUpperCase();
    })),
    slice: Utils.injectHelp(`String.slice(string, start = 0, end = String.sizeOf(string))`, createFunctionHandler(1, 3, (args, referrer, context) => {
        const string = args[0];
        if (typeof string !== 'string') {
            Utils.raise(TypeError, `expect a string as source`, referrer, context);
        }
        if (args.length > 1) {
            if (typeof args[1] !== 'number') {
                Utils.raise(TypeError, `expect a number as begin index`, referrer, context);
            }
            if (args.length > 2) {
                if (typeof args[2] !== 'number') {
                    Utils.raise(TypeError, `expect a number as end index`, referrer, context);
                }
            }
        }
        return string.slice(args[1], args[2]);
    })),
    indexOf: Utils.injectHelp(`String.indexOf(string, substring)`, createFunctionHandler(2, 2, (args, referrer, context) => {
        const string = args[0];
        if (typeof string !== 'string') {
            Utils.raise(TypeError, `expect a string to search in`, referrer, context);
        }
        const substring = args[1];
        if (typeof substring !== 'string') {
            Utils.raise(TypeError, `expect another string to search for`, referrer, context);
        }
        return string.indexOf(substring);
    })),
    lastIndexOf: Utils.injectHelp(`String.lastIndexOf(string, substring)`, createFunctionHandler(2, 2, (args, referrer, context) => {
        const string = args[0];
        if (typeof string !== 'string') {
            Utils.raise(TypeError, `expect a string to search in`, referrer, context);
        }
        const substring = args[1];
        if (typeof substring !== 'string') {
            Utils.raise(TypeError, `expect another string to search for`, referrer, context);
        }
        return string.lastIndexOf(substring);
    })),
    includes: Utils.injectHelp(`String.includes(string, substring)`, createFunctionHandler(2, 2, (args, referrer, context) => {
        const string = args[0];
        if (typeof string !== 'string') {
            Utils.raise(TypeError, `expect a string to search in`, referrer, context);
        }
        const substring = args[1];
        if (typeof substring !== 'string') {
            Utils.raise(TypeError, `expect another string to search for`, referrer, context);
        }
        return string.includes(substring);
    })),
    repeat: Utils.injectHelp(`String.repeat(string, count)`, createFunctionHandler(2, 2, (args, referrer, context) => {
        const string = args[0];
        if (typeof string !== 'string') {
            Utils.raise(TypeError, `expect a string to repeat`, referrer, context);
        }
        const count = args[1];
        if (typeof count !== 'number') {
            Utils.raise(TypeError, `expect a number as repeating count`, referrer, context);
        }
        if (!Number.isFinite(count) || count < 0) {
            Utils.raise(RangeError, `invalid count`, referrer, context);
        }
        return string.repeat(count);
    })),
    codePointAt: Utils.injectHelp(`String.codePointAt(string, index)`, createFunctionHandler(2, 2, (args, referrer, context) => {
        const string = args[0];
        if (typeof string !== 'string') {
            Utils.raise(TypeError, `expect a string`, referrer, context);
        }
        const index = args[1];
        if (typeof index !== 'number') {
            Utils.raise(TypeError, `expect a number as index`, referrer, context);
        }
        const normalizedIndex = Utils.normalizeIndex(index, string.length, referrer, context);
        const result = string.codePointAt(normalizedIndex);
        if (typeof result === 'number') {
            return result;
        }
        else {
            return null;
        }
    })),
}));

const miscOperators = [{
        symbol: '.',
        priority: 1,
        ltr: true,
        handler(buffer, index, context) {
            const target = evalBufferNode(buffer, index - 1, buffer[index], context);
            if (!Utils.isDict(target)) {
                Utils.raise(TypeError, 'expect a dict', buffer[index - 1], context);
            }
            const nameNode = buffer[index + 1];
            if (!nameNode || nameNode.type !== 'word') {
                Utils.raise(TypeError, 'expect a word following', buffer[index], context);
            }
            const name = nameNode.value;
            const value = getProperty(target, name, buffer[index + 1], buffer[index - 1], context);
            const valueNode = Utils.createValueNode(value, buffer[index]);
            Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
        },
    }, {
        symbol: ':',
        priority: 1,
        ltr: true,
        handler(buffer, index, context) {
            const target = evalBufferNode(buffer, index - 1, buffer[index], context);
            const nameNode = buffer[index + 1];
            if (!nameNode || nameNode.type !== 'word') {
                Utils.raise(TypeError, 'expect a word following', buffer[index], context);
            }
            let dict = null;
            switch (typeof target) {
                case 'number': {
                    dict = builtinNumber;
                    break;
                }
                case 'string': {
                    dict = builtinString;
                    break;
                }
                case 'boolean': {
                    dict = null;
                    break;
                }
                case 'function': {
                    dict = builtinFunction;
                    break;
                }
                case 'object': {
                    if (!target) {
                        dict = null;
                    }
                    else if (Array.isArray(target)) {
                        dict = builtinArray;
                    }
                    else if (Utils.isDict(target)) {
                        dict = builtinDict;
                    }
                    break;
                }
                default: {
                    dict = null;
                }
            }
            const name = nameNode.value;
            let callback;
            if (dict && name in dict) {
                callback = dict[name];
            }
            else {
                callback = null;
            }
            if (!isInvocable(callback)) {
                Utils.raise(TypeError, 'expect an invocable', buffer[index], context);
            }
            const COMMA_NODE = {
                type: 'symbol',
                value: ',',
                line: buffer[index].line,
                column: buffer[index].column,
                offset: buffer[index].offset,
            };
            const handler = (rawArgs, referrer, _context, thisArg) => {
                return invoke(callback, [
                    Utils.createValueNode(target, referrer),
                    COMMA_NODE,
                    ...rawArgs,
                ], referrer, _context, thisArg);
            };
            const valueNode = Utils.createValueNode(handler, buffer[index]);
            Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
        },
    }, {
        symbol: '??',
        priority: 12,
        ltr: true,
        handler: createBinaryOperator('__nullOr', null, null, (a, b) => (a ?? b)),
    }, {
        symbol: '#',
        priority: 2,
        ltr: true,
        handler(buffer, index, context) {
            const wordNode = buffer[index + 1];
            if (index + 1 === buffer.length || wordNode.type !== 'word') {
                Utils.raise(SyntaxError, 'expect a word following', buffer[index], context);
            }
            const valueNode = Utils.createValueNode(wordNode.value, buffer[index]);
            Utils.replaceBuffer(buffer, index, 2, valueNode);
        },
    }];

const parathesisHandler = (buffer, index, context) => {
    if (index === 0 || buffer[index - 1].type === 'symbol') { // pure paratheses
        const value = evalExpression(buffer[index].body, context);
        buffer[index] = Utils.createValueNode(value, buffer[index]);
    }
    else { // function call
        const handler = evalNode(buffer[index - 1], context);
        if (!isInvocable(handler)) {
            Utils.raise(TypeError, 'expect an invocable', buffer[index], context);
        }
        const value = invoke(handler, buffer[index].body, buffer[index], context, null);
        const valueNode = Utils.createValueNode(value, buffer[index]);
        Utils.replaceBuffer(buffer, index - 1, 2, valueNode);
    }
};

/**
 * List of operator definitions.
 */
const operators = [
    ...mathOperators,
    ...compareOperators,
    ...booleanOperators,
    ...bitOperators,
    ...assignmentOperators,
    ...miscOperators,
    {
        symbol: '@',
        priority: 0,
        ltr: true,
        handler: inlineFunctionHandler,
    },
    {
        symbol: '=>',
        priority: 0,
        ltr: true,
        handler: inlineExpressionHandler,
    },
    {
        symbol: '[',
        priority: 1,
        ltr: true,
        handler: braceHandler,
    },
    {
        symbol: '(',
        priority: 1,
        ltr: true,
        handler: parathesisHandler,
    },
    {
        symbol: '{',
        priority: 1,
        ltr: true,
        handler: bracketHandler,
    },
];
/**
 * A utility map. (operator->handler)
 */
const operatorHandlers = new Map(operators.map(op => [op.symbol, op.handler]));
/**
 * A utility map. (operator->priority)
 */
const operatorPriorities = new Map(operators.map(op => [op.symbol, op.priority]));
/**
 * A utility map. (operator->ltr)
 */
const operatorLtr = new Map(operators.map(op => [op.symbol, op.ltr]));

const e={normalSymbols:new Set("()[]{}<>+-*/\\?~!@#$%^&=|,.:;"),twoCharacterSymbols:new Set(["&&","||","??","==",">=","<=","!=","<<",">>","++","--","**","+=","-=","*=","/=","%=","&=","^=","|=","|>","->","=>","..","//","/*","*/"]),threeCharacterSymbols:new Set(["&&=","||=","??=","===","!==","<<=",">>=","**=","...",">>>"]),fourCharacterSymbols:new Set([">>>="]),globSymbols:new Set(['"',"'","`"]),numberCharacters:new Set("0123456789"),extendedNumberCharacters:new Set(".ABCDEF"),numberSuffixes:new Set(["D","B","O","H"]),spaceCharacters:new Set(" \r\n\t")},tokenize=(t,s)=>{const n=Object.assign({},e,s),{normalSymbols:a,twoCharacterSymbols:o,threeCharacterSymbols:l,fourCharacterSymbols:r,globSymbols:h,numberCharacters:c,extendedNumberCharacters:f,numberSuffixes:i,spaceCharacters:u}=n,m=[];let p=t[0],b="",g=0,S=!1;h.has(p)?b=p:u.has(p)?g=1:c.has(p)?g=2:a.has(p)&&(m.push(p),p="");for(let e=1;e<t.length;e++){const s=t[e];if(b){p+=s,S?S=!1:s===b?(m.push(p),p="",b="",g=0):S="\\"===s;continue}if(h.has(s)){p&&m.push(p),p=s,b=s,S=!1;continue}if((2===g||3===g)&&i.has(s)){p+=s,g=3;continue}if(0===g&&!p&&a.has(s)&&m.length){const e=m[m.length-1];if(1===e.length&&o.has(e+s)){m[m.length-1]+=s;continue}if(2===e.length&&l.has(e+s)){m[m.length-1]+=s;continue}if(3===e.length&&r.has(e+s)){m[m.length-1]+=s;continue}}let n=0;if(u.has(s))n=1;else if(c.has(s)||2===g&&f.has(s)&&!p.includes("."))n=2;else if(a.has(s)){p&&m.push(p),m.push(s),p="",g=0;continue}if(g===n)p+=s;else if(0===g&&2===n){if(p+=s,p.length>1)continue}else p&&m.push(p),p=s;g=n;}return p&&m.push(p),m},t=Object.assign({},e,{spanSymbols:new Map([["(",")"],["[","]"],["{","}"]]),escapeCharacters:new Map([["t","\t"],["r","\r"],["n","\n"],["\n",""],["\\","\\"],['"','"'],["'","'"],["`","`"]])}),token2ast=(e,s,n=0,a=1,o=1,l=0,r=e.length)=>{const h=Object.assign({},t,s),{normalSymbols:c,twoCharacterSymbols:f,threeCharacterSymbols:i,fourCharacterSymbols:u,globSymbols:m,numberCharacters:p,numberSuffixes:b,spaceCharacters:g,spanSymbols:S,escapeCharacters:y}=h,w=[];let C=n,x=a,d=o;for(let t=l;t<r;t++){const s=e[t];if(S.has(s)){const n=S.get(s);let a=0,o=t+1;for(;o<e.length;o++)if(e[o]!==s){if(e[o]===n){if(!a)break;a--;}}else a++;if(o===e.length)throw new SyntaxError(`missing ${n} for ${s} at line ${x} column ${d}`);if(o===t+1)w.push({type:"span",start:s,stop:n,body:[],offset:C,line:x,column:d}),C++;else {const a=token2ast(e,h,C+1,x,d+1,t+1,o),l=a.ast;w.push({type:"span",start:s,stop:n,body:l,offset:C,line:x,column:d}),C=a.stopOffset,x=a.stopLine,d=a.stopColumn;}t=o;}else if(c.has(s)||f.has(s)||i.has(s)||u.has(s))w.push({type:"symbol",value:s,offset:C,line:x,column:d});else if(m.has(s[0])){let e=s[0],t=!1;for(let n=1;n<s.length-1;n++){const a=s[n];if(t){if(!y.has(a)){if("\r"===a&&"\n"===s[n+1])continue;throw new SyntaxError("unknown escape character: \\"+a)}e+=y.get(a),t=!1;}else {if("\\"===a){t=!0;continue}e+=a;}}if(t||s[s.length-1]!==s[0]||1===s.length)throw new SyntaxError(`missing corresponding ${S} at line ${x} column ${d}`);e+=s[0],w.push({type:"glob",value:e,offset:C,line:x,column:d});}else if(p.has(s[0])){let e=s.length-1;for(;e>=0&&b.has(s[e]);e--);e++,w.push({type:"number",value:s.slice(0,e),suffix:s.slice(e),offset:C,line:x,column:d});}else g.has(s[0])||w.push({type:"word",value:s,offset:C,line:x,column:d});C+=s.length;for(let e=0;e<s.length;e++)"\n"===s[e]&&x++;const n=s.lastIndexOf("\n");~n?d=s.length-n:d+=s.length;}return {ast:w,stopOffset:C,stopLine:x,stopColumn:d}},parse=(e,t)=>token2ast(tokenize(e,t),t);

const createIf = (ended) => Utils.injectHelp('if (condition0) { ... } (condition1) { ... } ...', createFunctionHandler(1, 1, (args, referrer, context) => {
    const condition = args[0];
    if (typeof condition !== 'boolean') {
        Utils.raise(TypeError, 'expect a boolean as condition', referrer, context);
    }
    return createFunctionHandler(1, 1, (_args, referrer, context) => {
        const callback = _args[0];
        if (!isInvocable(callback)) {
            Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
        }
        if (!ended && condition) {
            invoke(callback, [], referrer, context, null);
            return createIf(true);
        }
        else {
            return createIf(false);
        }
    });
}));
const builtinIf = createIf(false);

const builtinFor = Utils.injectHelp('for (iteratorName, begin, end, step?) { ... }', createFunctionHandler(3, 4, (args, referrer, context) => {
    const iteratorName = args[0];
    if (typeof iteratorName !== 'string') {
        Utils.raise(TypeError, 'expect a string as iterator name', referrer, context);
    }
    const begin = args[1];
    if (!Number.isFinite(begin)) {
        Utils.raise(TypeError, 'expect a finite number as begin value', referrer, context);
    }
    const end = args[2];
    if (!Number.isFinite(end)) {
        Utils.raise(TypeError, 'expect a finite number as end value', referrer, context);
    }
    const defaultStep = begin <= end ? 1 : -1;
    const step = args.length === 4
        ? args[3]
        : defaultStep;
    if (!Number.isFinite(step)) {
        Utils.raise(TypeError, 'expect a finite number as step', referrer, context);
    }
    if (begin !== end
        && Math.sign(defaultStep) !== Math.sign(step)) {
        Utils.raise(RangeError, 'invalid step', referrer, context);
    }
    return createFunctionHandler(1, 1, (_args, _referrer, _context) => {
        const callback = _args[0];
        if (!isInvocable(callback)) {
            Utils.raise(TypeError, 'expect an invocable as callback', _referrer, _context);
        }
        const BREAK_FLAG = Symbol('hxs-break-flag');
        const CONTINUE_FLAG = Symbol('hxs-break-flag');
        Utils.injectTemp(_context.store, {
            break: createFunctionHandler(0, 0, () => {
                throw BREAK_FLAG;
            }),
            continue: createFunctionHandler(0, 0, () => {
                throw CONTINUE_FLAG;
            }),
        }, () => {
            if (step > 0) {
                for (let iterator = begin; iterator < end; iterator += step) {
                    try {
                        _context.store[iteratorName] = iterator;
                        invoke(callback, [], referrer, _context, null);
                    }
                    catch (error) {
                        if (error === BREAK_FLAG) {
                            break;
                        }
                        else if (error !== CONTINUE_FLAG) {
                            throw error;
                        }
                    }
                }
            }
            else {
                for (let iterator = begin; iterator > end; iterator += step) {
                    try {
                        _context.store[iteratorName] = iterator;
                        invoke(callback, [], referrer, _context, null);
                    }
                    catch (error) {
                        if (error === BREAK_FLAG) {
                            break;
                        }
                        else if (error !== CONTINUE_FLAG) {
                            throw error;
                        }
                    }
                }
            }
        });
        return null;
    });
}));

const builtinWhile = Utils.injectHelp('while (condition) { ... }', (rawArgs, referrer, context) => (createFunctionHandler(1, 1, (_args, _referrer, _context) => {
    const callback = _args[0];
    if (!isInvocable(callback)) {
        Utils.raise(TypeError, 'expect an invocable as callback', _referrer, _context);
    }
    const BREAK_FLAG = Symbol('hxs-break-flag');
    const CONTINUE_FLAG = Symbol('hxs-break-flag');
    Utils.injectTemp(_context.store, {
        break: createFunctionHandler(0, 0, () => {
            throw BREAK_FLAG;
        }),
        continue: createFunctionHandler(0, 0, () => {
            throw CONTINUE_FLAG;
        }),
    }, () => {
        while (true) {
            const condition = evalExpression(rawArgs, context);
            if (typeof condition !== 'boolean') {
                Utils.raise(TypeError, 'expect a boolean as condition', referrer, context);
            }
            if (!condition) {
                break;
            }
            try {
                invoke(callback, [], referrer, _context, null);
            }
            catch (error) {
                if (error === BREAK_FLAG) {
                    break;
                }
                else if (error !== CONTINUE_FLAG) {
                    throw error;
                }
            }
        }
    });
    return null;
})));

const builtinMath = Utils.injectHelp('A dict providing constants and methods for math calculation.', Utils.createDict({
    PI: Math.PI,
    E: Math.E,
    random: Utils.injectHelp('Math.random()', createFunctionHandler(0, 0, Math.random)),
    sign: Utils.injectHelp('Math.sign(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.sign(args[0]);
    })),
    floor: Utils.injectHelp('Math.floor(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.floor(args[0]);
    })),
    ceil: Utils.injectHelp('Math.ceil(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.ceil(args[0]);
    })),
    round: Utils.injectHelp('Math.round(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.round(args[0]);
    })),
    sqrt: Utils.injectHelp('Math.sqrt(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const x = args[0];
        if (typeof x !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        if (x < 0) {
            Utils.raise(RangeError, `invalid radicant`, referrer, context);
        }
        return Math.sqrt(x);
    })),
    min: Utils.injectHelp('Math.min(numbers)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const numbers = args[0];
        if (!Array.isArray(numbers)) {
            Utils.raise(TypeError, `expect an array of numbers`, referrer, context);
        }
        let min = Infinity;
        for (let i = 0; i < numbers.length; i++) {
            if (typeof numbers[i] !== 'number' || numbers[i] !== numbers[i]) {
                Utils.raise(TypeError, `invalid input`, referrer, context);
            }
            if (numbers[i] < min) {
                min = numbers[i];
            }
        }
        return min;
    })),
    max: Utils.injectHelp('Math.max(numbers)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const numbers = args[0];
        if (!Array.isArray(numbers)) {
            Utils.raise(TypeError, `expect an array of numbers`, referrer, context);
        }
        let max = -Infinity;
        for (let i = 0; i < numbers.length; i++) {
            if (typeof numbers[i] !== 'number' || numbers[i] !== numbers[i]) {
                Utils.raise(TypeError, `invalid input`, referrer, context);
            }
            if (numbers[i] > max) {
                max = numbers[i];
            }
        }
        return max;
    })),
    sum: Utils.injectHelp('Math.sum(numbers)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const numbers = args[0];
        if (!Array.isArray(numbers)) {
            Utils.raise(TypeError, `expect an array of numbers`, referrer, context);
        }
        let sum = 0;
        for (let i = 0; i < numbers.length; i++) {
            if (typeof numbers[i] !== 'number' || numbers[i] !== numbers[i]) {
                Utils.raise(TypeError, `invalid input`, referrer, context);
            }
            sum += numbers[i];
        }
        return sum;
    })),
    mean: Utils.injectHelp('Math.mean(numbers)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const numbers = args[0];
        if (!Array.isArray(numbers)) {
            Utils.raise(TypeError, `expect an array of numbers`, referrer, context);
        }
        if (!numbers.length) {
            Utils.raise(RangeError, `no numbers given`, referrer, context);
        }
        let sum = 0;
        for (let i = 0; i < numbers.length; i++) {
            if (typeof numbers[i] !== 'number' || numbers[i] !== numbers[i]) {
                Utils.raise(TypeError, `invalid input`, referrer, context);
            }
            sum += numbers[i];
        }
        return sum / numbers.length;
    })),
    log: Utils.injectHelp('Math.log(a, b)', createFunctionHandler(2, 2, (args, referrer, context) => {
        const a = args[0];
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as base`, referrer, context);
        }
        if (a <= 0 || a === 1 || a === Infinity) {
            Utils.raise(RangeError, `invalid base`, referrer, context);
        }
        const b = args[1];
        if (typeof b !== 'number') {
            Utils.raise(TypeError, `expect a number as natural number`, referrer, context);
        }
        if (b <= 0) {
            Utils.raise(RangeError, `invalid natural number`, referrer, context);
        }
        return Math.log(b) / Math.log(a);
    })),
    ln: Utils.injectHelp('Math.ln(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const x = args[0];
        if (typeof x !== 'number') {
            Utils.raise(TypeError, `expect a number as natural number`, referrer, context);
        }
        if (x <= 0) {
            Utils.raise(RangeError, `invalid natural number`, referrer, context);
        }
        return Math.log(x);
    })),
    lg: Utils.injectHelp('Math.lg(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const x = args[0];
        if (typeof x !== 'number') {
            Utils.raise(TypeError, `expect a number as natural number`, referrer, context);
        }
        if (x <= 0) {
            Utils.raise(RangeError, `invalid natural number`, referrer, context);
        }
        return Math.log(x) / Math.LN10;
    })),
    sin: Utils.injectHelp('Math.sin(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.sin(args[0]);
    })),
    cos: Utils.injectHelp('Math.cos(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.cos(args[0]);
    })),
    tan: Utils.injectHelp('Math.tan(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.tan(args[0]);
    })),
    csc: Utils.injectHelp('Math.csc(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return 1 / Math.sin(args[0]);
    })),
    sec: Utils.injectHelp('Math.sec(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return 1 / Math.cos(args[0]);
    })),
    cot: Utils.injectHelp('Math.cot(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return 1 / Math.tan(args[0]);
    })),
    asin: Utils.injectHelp('Math.asin(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.asin(args[0]);
    })),
    acos: Utils.injectHelp('Math.acos(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.acos(args[0]);
    })),
    atan: Utils.injectHelp('Math.atan(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.atan(args[0]);
    })),
    acsc: Utils.injectHelp('Math.acsc(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.asin(1 / args[0]);
    })),
    asec: Utils.injectHelp('Math.asec(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.acos(1 / args[0]);
    })),
    acot: Utils.injectHelp('Math.acot(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.atan(1 / args[0]);
    })),
    sinh: Utils.injectHelp('Math.sinh(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.sinh(args[0]);
    })),
    cosh: Utils.injectHelp('Math.cosh(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.cosh(args[0]);
    })),
    tanh: Utils.injectHelp('Math.tanh(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.tanh(args[0]);
    })),
    csch: Utils.injectHelp('Math.csch(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return 1 / Math.sinh(args[0]);
    })),
    sech: Utils.injectHelp('Math.sech(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return 1 / Math.cosh(args[0]);
    })),
    coth: Utils.injectHelp('Math.coth(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return 1 / Math.tanh(args[0]);
    })),
    asinh: Utils.injectHelp('Math.asinh(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.asinh(args[0]);
    })),
    acosh: Utils.injectHelp('Math.acosh(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.acosh(args[0]);
    })),
    atanh: Utils.injectHelp('Math.atanh(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.atanh(args[0]);
    })),
    acsch: Utils.injectHelp('Math.acsch(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.asinh(1 / args[0]);
    })),
    asech: Utils.injectHelp('Math.asech(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.acosh(1 / args[0]);
    })),
    acoth: Utils.injectHelp('Math.acoth(x)', createFunctionHandler(1, 1, (args, referrer, context) => {
        if (typeof args[0] !== 'number') {
            Utils.raise(TypeError, `expect a number as argument`, referrer, context);
        }
        return Math.atanh(1 / args[0]);
    })),
}));

const typeOf = Utils.injectHelp('typeOf(value)', createFunctionHandler(1, 1, (args, referrer, context) => {
    const value = args[0];
    switch (typeof value) {
        case 'number': {
            return 'number';
        }
        case 'string': {
            return 'string';
        }
        case 'boolean': {
            return 'boolean';
        }
        case 'function': {
            return 'function';
        }
        case 'object': {
            if (!value) {
                return 'null';
            }
            else if (Array.isArray(value)) {
                return 'array';
            }
            else if (Utils.isDict(value)) {
                return 'dict';
            }
        }
        default: {
            Utils.raise(TypeError, 'unrecognized type', referrer, context);
            return null;
        }
    }
}));
const number = Utils.injectHelp('number(value)', createFunctionHandler(1, 1, (args, referrer, context) => {
    const type = typeof args[0];
    if (type === 'object' || type === 'function') {
        return NaN;
    }
    else {
        return Number(args[0]);
    }
}));
const string = Utils.injectHelp(`string(value, previewSize = 10, previewIndent = '  ')`, createFunctionHandler(1, 3, (args, referrer, context) => {
    const previewSize = args.length >= 2 ? args[1] : 10;
    if (typeof previewSize !== 'number') {
        Utils.raise(TypeError, 'expect a number as preview size', referrer, context);
    }
    const previewIndent = args.length >= 3 ? args[2] : '  ';
    if (typeof previewIndent !== 'string') {
        Utils.raise(TypeError, 'expect a string as preview indent', referrer, context);
    }
    return Utils.toString(args[0], previewSize, previewIndent);
}));
const boolean = Utils.injectHelp('boolean(value)', createFunctionHandler(1, 1, (args, referrer, context) => {
    return Boolean(args[0]);
}));

const help = Utils.injectHelp('help(target)', createFunctionHandler(1, 1, (args, referrer, context) => {
    const target = args[0];
    switch (typeof target) {
        case 'function':
        case 'object': {
            if (target && HELP_SYMBOL in target) {
                return target[HELP_SYMBOL];
            }
            break;
        }
    }
    return 'no help available';
}));
const injectHelp = Utils.injectHelp('injectHelp(helpInfo, target)', createFunctionHandler(2, 2, (args, referrer, context) => {
    const helpInfo = args[0];
    if (typeof helpInfo !== 'string') {
        Utils.raise(TypeError, 'expect a string as help info', referrer, context);
    }
    const target = args[1];
    if (typeof target !== 'function' && !Utils.isDict(target)) {
        Utils.raise(TypeError, 'expect a dict or function as target', referrer, context);
    }
    target[HELP_SYMBOL] = helpInfo;
    return target;
}));

const exist = Utils.injectHelp('exist(variableName)', createFunctionHandler(1, 1, (args, referrer, context) => {
    const name = args[0];
    if (typeof name !== 'string') {
        Utils.raise(TypeError, 'expect a string as variable name', referrer, context);
    }
    return name in context.store;
}));
const delete_ = Utils.injectHelp('delete(variableName)', createFunctionHandler(1, 1, (args, referrer, context) => {
    const name = args[0];
    if (typeof name !== 'string') {
        Utils.raise(TypeError, 'expect a string as variable name', referrer, context);
    }
    const { store } = context;
    if (!(name in store)) {
        Utils.raise(ReferenceError, `${Utils.toDisplay(name)} is not defined`, referrer, context);
    }
    delete store[name];
    return null;
}));
const getContextStore = Utils.injectHelp('getContextStore()', createFunctionHandler(0, 0, (args, referrer, context) => {
    return context.store;
}));

const print = Utils.injectHelp('print(message...)', createFunctionHandler(0, Infinity, (args, referrer, context) => {
    let message = '';
    for (let i = 0; i < args.length; i++) {
        message += Utils.toString(args[i]);
    }
    console.log(message);
    return null;
}));
const same = Utils.injectHelp('same(a, b)', createFunctionHandler(2, 2, (args, referrer, context) => {
    return Object.is(args[0], args[1]);
}));
const keys = Utils.injectHelp('keys(dict)', createFunctionHandler(1, 1, (args, referrer, context) => {
    const dict = args[0];
    if (!Utils.isDict(dict)) {
        Utils.raise(TypeError, 'expect a dict', referrer, context);
    }
    return getKeysOf(dict, referrer, context);
}));
const remove = Utils.injectHelp('remove(arrayOrDict, indexOrKey)', createFunctionHandler(2, 2, (args, referrer, context) => {
    removeProperty(args[0], args[1], referrer, referrer, context);
    return null;
}));
const sizeOf = Utils.injectHelp('sizeOf(arrayOrString)', createFunctionHandler(1, 1, (args, referrer, context) => {
    const target = args[0];
    if (Array.isArray(target) || typeof target === 'string') {
        return target.length;
    }
    else {
        Utils.raise(TypeError, `expect an array or string`, referrer, context);
        return null; // for type checking
    }
}));
const slice = Utils.injectHelp('slice(arrayOrString, begin = 0, end = sizeOf(arrayOrString))', createFunctionHandler(1, 3, (args, referrer, context) => {
    const target = args[0];
    if (Array.isArray(target) || typeof target === 'string') {
        if (args.length > 1) {
            if (typeof args[1] !== 'number') {
                Utils.raise(TypeError, 'expect a number as begin index', referrer, context);
            }
            if (args.length > 2) {
                if (typeof args[2] !== 'number') {
                    Utils.raise(TypeError, 'expect a number as end index', referrer, context);
                }
            }
        }
        return target.slice(args[1], args[2]);
    }
    else {
        Utils.raise(TypeError, `expect an array or string`, referrer, context);
        return null; // for type checking
    }
}));
const clone = Utils.injectHelp('clone(arrayOrDict)', createFunctionHandler(1, 1, (args, referrer, context) => {
    const target = args[0];
    if (Array.isArray(target)) {
        return target.slice();
    }
    else if (Utils.isDict(target)) {
        return Utils.createDict(target);
    }
    else {
        Utils.raise(TypeError, `expect an array or dict to clone`, referrer, context);
        return null; // for type checking
    }
}));
const indexOf = Utils.injectHelp('indexOf(arrayOrString, valueOrSubstring)', createFunctionHandler(2, 2, (args, referrer, context) => {
    const target = args[0];
    const targetIsString = typeof target === 'string';
    if (!targetIsString && !Array.isArray(target)) {
        Utils.raise(TypeError, 'expect an array to search in', referrer, context);
    }
    if (targetIsString && typeof args[1] !== 'string') {
        Utils.raise(TypeError, 'expect another string to search for', referrer, context);
    }
    return target.indexOf(args[1]);
}));
const lastIndexOf = Utils.injectHelp('lastIndexOf(arrayOrString, valueOrSubstring)', createFunctionHandler(2, 2, (args, referrer, context) => {
    const target = args[0];
    const targetIsString = typeof target === 'string';
    if (!targetIsString && !Array.isArray(target)) {
        Utils.raise(TypeError, 'expect an array to search in', referrer, context);
    }
    if (targetIsString && typeof args[1] !== 'string') {
        Utils.raise(TypeError, 'expect another string to search for', referrer, context);
    }
    return target.lastIndexOf(args[1]);
}));
const includes = Utils.injectHelp('includes(arrayOrString, valueOrSubstring)', createFunctionHandler(2, 2, (args, referrer, context) => {
    const target = args[0];
    const targetIsString = typeof target === 'string';
    if (!targetIsString && !Array.isArray(target)) {
        Utils.raise(TypeError, 'expect an array to search in', referrer, context);
    }
    if (targetIsString && typeof args[1] !== 'string') {
        Utils.raise(TypeError, 'expect another string to search for', referrer, context);
    }
    return target.includes(args[1]);
}));

const builtinRaise = Utils.injectHelp('raise(error_message)', createFunctionHandler(1, 1, (args, referrer, context) => {
    const message = args[0];
    if (typeof message !== 'string') {
        Utils.raise(TypeError, 'expect a string as error message', referrer, context);
    }
    Utils.raise(Error, message, referrer, context);
    return null;
}));
const builtinTry = Utils.injectHelp('try { ... } (error) { ... }', createFunctionHandler(1, 1, (args, referrer, context, thisArg) => {
    const tryBody = args[0];
    if (!isInvocable(tryBody)) {
        Utils.raise(TypeError, 'expect an invocable to try', referrer, context);
    }
    // catch
    return (rawArgs) => {
        if (rawArgs.length !== 1 || rawArgs[0].type !== 'word') {
            Utils.raise(TypeError, 'expect a word as the name for error', referrer, context);
        }
        const errName = rawArgs[0].value;
        return createFunctionHandler(1, 1, (_args, _referrer, _context, _thisArg) => {
            const catchBody = _args[0];
            if (!isInvocable(catchBody)) {
                Utils.raise(TypeError, 'expect an invocable as error handler', referrer, context);
            }
            const tryStackLength = context.stack.length;
            try {
                invoke(tryBody, [], referrer, context, thisArg);
            }
            catch (error) {
                if (typeof error === 'symbol') {
                    throw error;
                }
                Utils.injectTemp(_context.store, {
                    [errName]: String(error),
                }, () => {
                    context.stack.length = tryStackLength;
                    invoke(catchBody, [], _referrer, _context, _thisArg);
                });
            }
            return null;
        });
    };
}));

const builtinClass = Utils.injectHelp('A dict providing class APIs.', Utils.createDict({
    __invoke: Utils.injectHelp('Class.__invoke(description)', createFunctionHandler(1, 1, (args, referrer, context) => {
        const description = args[0];
        if (!Utils.isDict(description)) {
            Utils.raise(TypeError, 'expect a dict as class description', referrer, context);
        }
        if (('__init' in description) && !isInvocable(description.__init)) {
            Utils.raise(TypeError, 'expect an invocable as initializer', referrer, context);
        }
        return createClass(description, referrer, context);
    })),
    extend: Utils.injectHelp('Class.extend(baseClass, baseArgs = [])(description)', createFunctionHandler(1, 2, (args, referrer, context) => {
        const baseClass = args[0];
        if (!isInvocable(baseClass)) {
            Utils.raise(TypeError, 'expect an invocable as base class', referrer, context);
        }
        const baseArgs = args.length > 1 ? args[1] : [];
        if (!Array.isArray(baseArgs)) {
            Utils.raise(TypeError, 'expect an array as base args', referrer, context);
        }
        return createFunctionHandler(1, 1, (_args, _referrer, _context) => {
            const description = _args[0];
            if (!Utils.isDict(description)) {
                Utils.raise(TypeError, 'expect a dict as class description', referrer, context);
            }
            if (('__init' in description) && !isInvocable(description.__init)) {
                Utils.raise(TypeError, 'expect an invocable as initializer', referrer, context);
            }
            if ('__init' in description) {
                const oldInit = description.__init;
                const newInit = (rawArgs, ref, ctx, thisArg) => {
                    invoke(baseClass, Utils.createRawArray(baseArgs, ref), referrer, context, thisArg);
                    return invoke(oldInit, rawArgs, ref, ctx, thisArg);
                };
                description.__init = newInit;
            }
            else {
                const init = (rawArgs, ref, ctx, thisArg) => {
                    invoke(baseClass, Utils.createRawArray(baseArgs, ref), referrer, context, thisArg);
                    return null;
                };
                description.__init = init;
            }
            const constructor = createClass(description, _referrer, _context);
            constructor[BASE_SYMBOL] = baseClass;
            return constructor;
        });
    })),
}));

const getExports = Utils.injectHelp('getExports()', createFunctionHandler(0, 0, (args, referrer, context) => {
    return context.exports;
}));
const export_ = Utils.injectHelp('export(exports)', createFunctionHandler(1, 1, (args, referrer, context) => {
    const exports = args[0];
    if (!Utils.isDict(exports)) {
        Utils.raise(TypeError, `expect a dict as exports`, referrer, context);
    }
    Object.assign(context.exports, exports);
    return null;
}));

const builtinJSON = Utils.injectHelp('A dict providing methods for JSON operations.', Utils.createDict({
    parse: Utils.injectHelp(`JSON.parse(string)`, createFunctionHandler(1, 1, (args, referrer, context) => {
        const string = args[0];
        if (typeof string !== 'string') {
            Utils.raise(TypeError, `expect a string to parse`, referrer, context);
        }
        return JSON.parse(string);
    })),
    stringify: Utils.injectHelp(`JSON.stringify(value, space = '')`, createFunctionHandler(1, 2, (args, referrer, context) => {
        const space = (args.length > 1) ? args[1] : '';
        if (typeof space !== 'string') {
            Utils.raise(TypeError, `expect a string as space argument`, referrer, context);
        }
        return JSON.stringify(args[0], undefined, space);
    })),
}));

const builtins = Utils.createDict({
    true: true,
    false: false,
    Infinity,
    NaN,
    null: null,
    _: null,
    if: builtinIf,
    while: builtinWhile,
    for: builtinFor,
    raise: builtinRaise,
    try: builtinTry,
    Class: builtinClass,
    Array: builtinArray,
    String: builtinString,
    Dict: builtinDict,
    Function: builtinFunction,
    Number: builtinNumber,
    Math: builtinMath,
    JSON: builtinJSON,
    help,
    injectHelp,
    exist,
    delete: delete_,
    getContextStore,
    typeOf,
    number,
    string,
    boolean,
    print,
    same,
    keys,
    remove,
    sizeOf,
    slice,
    clone,
    indexOf,
    lastIndexOf,
    includes,
    getExports,
    export: export_,
    forEach: builtinArray.forEach,
    map: builtinArray.map,
    filter: builtinArray.filter,
    assign: builtinDict.assign,
    getConstructorOf: builtinDict.getConstructorOf,
    isInstanceOf: builtinDict.isInstanceOf,
    isInvocable: builtinFunction.isInvocable,
    invoke: builtinFunction.invoke,
    bind: builtinFunction.bind,
    isNaN: builtinNumber.isNaN,
});

/**
 * Execute the given code.
 */
const evalCode = (code, context) => {
    const _context = {};
    if (context) {
        Object.assign(_context, context);
    }
    if (!_context.store) {
        _context.store = Utils.createDict(builtins);
    }
    if (!_context.exports) {
        _context.exports = Object.create(null);
    }
    if (!_context.resolvedModules) {
        _context.resolvedModules = Object.create(null);
    }
    if (!_context.source) {
        _context.source = 'unknown';
    }
    if (!_context.basePath) {
        _context.basePath = '';
    }
    if (!_context.stack) {
        _context.stack = [];
    }
    return evalNodes(parse(code).ast, _context);
};

export { BASE_SYMBOL, CONSTRUCTOR_SYMBOL, HELP_SYMBOL, Utils, builtins, compileExpression, compileNodes, createAdditionalAssignmentOperator, createBinaryOperator, createClass, createFunctionHandler, createInlineExpression, createInlineFunction, createUnaryOperator, evalBinaryOperation, evalBufferNode, evalCode, evalCompiledExpression, evalCompiledNodes, evalExpression, evalList, evalNode, evalNodes, evalUnaryOperation, getConstructorOf, getKeysOf, getOperatorSymbol, getProperty, hasProperty, inlineExpressionHandler, inlineFunctionHandler, invoke, isInstanceOf, isInvocable, operatorHandlers, operatorLtr, operatorPriorities, operators, parseArg, parseArgList, removeProperty, setProperty };
